const global = window; const process = { nextTick: cb => new Promise(() => cb()) }

import randomBytes from 'randombytes';
import EventSource from 'eventsource';
import fetch from 'isomorphic-fetch';
import require$$0 from 'debug';
import isBuffer from 'is-buffer';
import base32Encode from 'base32-encode';
import require$$1 from 'streamx';
import require$$2 from 'events';
import require$$3 from '@arsonar/common';
import parseUrl from 'parse-dat-url';

const debug$2 = require$$0('sonar:fetch');


/**
  * Fetch a resource.
  *
  * This is a wrapper around the fetch web API. It should be API compatible to fetch,
  * with the following changes:
  *
  * @async
  * @param {string} [opts.endpoint=''] Endpoint URL (will be prefixed to URL)
  * @param {string} [opts.requestType='json'] Request encoding and content type.
  *   Supported values are 'json' and 'binary'
  * @param {string} [opts.responseType='text'] Response encoding. If the response
  *    has a JSON content type, will always be set to 'json'.
  *    Supported values are 'text', 'binary' and 'stream'.
  * @param {object} [opts.params] Query string parameters (will be encoded correctly).
  *
  * @return {Promise<object>} If the response has a JSON content type header, the
  *    decoded JSON will be returned. if opts.responseType is 'binary' or 'text',
  *    the response will be returned as a buffer or text.
  *
  * TODO: Rethink the default responseType cascade.
  */
var fetch_1 = async function makeFetch (url, opts) {
  if (!url.match(/^https?:\/\//)) {
    if (url.indexOf('://') !== -1) throw new Error('Only http: and https: protocols are supported.')
    if (!url.startsWith('/')) url = '/' + url;
    if (opts.endpoint) url = opts.endpoint + url;
  }

  if (!opts.headers) opts.headers = {};
  if (!opts.requestType) {
    if (isBuffer(opts.body)) opts.requestType = 'buffer';
    else opts.requestType = 'json';
  }

  if (opts.params) {
    const searchParams = new URLSearchParams();
    for (const [key, value] of Object.entries(opts.params)) {
      searchParams.append(key, value);
    }
    url += '?' + searchParams.toString();
  }

  if (opts.requestType === 'json') {
    opts.body = JSON.stringify(opts.body);
    opts.headers['content-type'] = 'application/json';
  }
  if (opts.requestType === 'buffer') {
    opts.headers['content-type'] = 'application/octet-stream';
  }

  const log = opts.log === false ? () => {} : (opts.log || debug$2);

  try {
    debug$2('fetch', url, opts);
    const res = await fetch(url, opts);
    if (!res.ok) {
      let message;
      if (isJsonResponse(res)) {
        message = (await res.json()).error;
      } else {
        message = await res.text();
      }
      log(`error ${res.status} ${url}`);
      throw new Error('Remote error (code ' + res.status + '): ' + message)
    }

    log(`ok ${res.status} fetch ${url}`);

    if (opts.responseType === 'stream') {
      return res.body
    }
    if (opts.responseType === 'buffer') {
      // nodejs only: res.buffer() returns a Buffer instance.
      if (res.buffer) return await res.buffer()
      // browser: Fetch API res.arrayBuffer returns ArrayBuffer.
      else return await res.arrayBuffer()
    }

    if (isJsonResponse(res)) {
      return await res.json()
    }

    return await res.text()
  } catch (err) {
    // TODO: If error fails for insufficient authorization, try creating
    // a new token if accessCode is set
    debug$2('fetch error', err);
    log(`error: fetch ${url} ${err.message}`);
    throw err
  }
};

function isJsonResponse (res) {
  const header = res.headers.get('content-type');
  if (!header) return false
  return header.indexOf('application/json') !== -1
}

var DEFAULT_ENDPOINT$1 = 'http://localhost:9191/api';
var DEFAULT_WORKSPACE$1 = 'default';
var DEFAULT_COLLECTION$1 = 'default';
var SCHEMA_RESOURCE$1 = 'sonar/resource';
var METADATA_ID$1 = 'sonar.id';
var HYPERDRIVE_SCHEME$2 = 'hyper:';

var constants = {
	DEFAULT_ENDPOINT: DEFAULT_ENDPOINT$1,
	DEFAULT_WORKSPACE: DEFAULT_WORKSPACE$1,
	DEFAULT_COLLECTION: DEFAULT_COLLECTION$1,
	SCHEMA_RESOURCE: SCHEMA_RESOURCE$1,
	METADATA_ID: METADATA_ID$1,
	HYPERDRIVE_SCHEME: HYPERDRIVE_SCHEME$2
};

const { HYPERDRIVE_SCHEME: HYPERDRIVE_SCHEME$1 } = constants;


/**
 * File system for a collection.
 */
class Fs {
  /**
   * File system for a collection.
   *
   * @constructor
   * @param {Collection} collection - Collection
   */
  constructor (collection) {
    this.endpoint = collection.endpoint + '/fs';
    this.collection = collection;
  }

  async fetch (path, opts = {}) {
    path = this.resolveURL(path);
    return this.collection.fetch(path, opts)
  }

  /**
   * Resolve a file URL into a HTTP url.
   *
   * @param {string} path - Either a hyper:// URL or a path in the collection's file system.
   * @throws Throws if the URL cannot be resolved.
   * @return {string} The HTTP URL to the file.
   */
  resolveURL (path) {
    // Support hyper:// URLs.
    if (path.startsWith(HYPERDRIVE_SCHEME$1)) {
      const url = parseUrl(path);
      path = url.host + url.path;
    // Support no other absolute URLs for now.
    } else if (path.indexOf('://') !== -1) {
      throw new Error('Invalid path: Only hyper:// URLs or paths are supported')
    }
    // Assume it's a path to a file in the island fs, will fail if not found.
    if (!path.startsWith('/')) path = '/' + path;
    return this.endpoint + path
  }

  /**
   * List the drives that are part of this collection.
   *
   * @async
   * @return {Promise<Array<object>>} Array of drive objects with keys `{ alias, key, writable }`
   */
  async listDrives () {
    // TODO: Move route under /fs somehow. Maybe HEAD on /?
    return this.collection.fetch('/fs-info')
  }

  /**
   * Get the contents of a directory.
   *
   * @async
   * @param {string} path - A hyper:// URL or a relative path within the collection's file system.
   * @param {object} [opts] - Options
   * @param {string} [opts.includeStats=true] - Include metadata for each file
   * @throws Will throw if the path is not found or is not a directory.
   * @return {Promise<Array<object>>} An array of objects with file metadata.
   */
  async readdir (path, opts = {}) {
    const self = this;
    path = path || '/';
    if (path === '/') {
      const drives = await this.listDrives();
      return drives.map(drive => ({
        name: drive.alias || drive.key,
        link: '',
        resource: null,
        length: null,
        directory: true
      }))
    }
    if (path.length > 2 && path.charAt(0) === '/') path = path.substring(1);
    const alias = path.split('/')[0];

    let files = await this.fetch(path);

    if (files && files.length) {
      files = files.map(file => {
        file.link = makeLink(file);
        file.resource = getResource(file);
        return file
      });
    }
    return files

    function makeLink (file) {
      return `${self.endpoint}/${self.collection.name}/fs/${alias}/${file.path}`
    }

    function getResource (file) {
      if (!file || !file.metadata || !file.metadata['sonar.id']) return null
      return file.metadata['sonar.id']
    }
  }

  /**
   * Write a file
   *
   * @async
   * @param {string} path - A hyper:// URL or a relative path within the collection's file system.
   * @param {Stream|Buffer} file - File content to write. Can be either a buffer or a read stream.
   * @param {object} [opts] - Options.
   * @throws Will throw if the path cannot be written to.
   * @return {Promise}
   */
  async writeFile (path, file, opts = {}) {
    const requestType = opts.requestType || 'buffer';
    const params = {};
    if (opts.metadata) params.metadata = JSON.stringify(opts.metadata);

    return this.fetch(path, {
      method: 'PUT',
      body: file,
      params,
      responseType: 'text',
      requestType,
      onUploadProgress: opts.onUploadProgress
    })
  }

  /**
   * Read a file into a buffer.
   *
   * @async
   * @param {string} path - A hyper:// URL or a relative path within the collection's file system.
   * @param {object} [opts] - Options. TODO: document.
   * @throws Will throw if the path is not found.
   * @return {Promise<ArrayBuffer|Buffer>} The file content. A Buffer object in Node.js, a ArrayBuffer object in the browser.
   */
  async readFile (path, opts = {}) {
    opts.responseType = opts.responseType || 'buffer';
    opts.requestType = opts.requestType || 'buffer';
    return this.fetch(path, opts)
  }

  /**
   * Get a read stream for a file.
   *
   * @async
   * @param {string} path - A hyper:// URL or a relative path within the collection's file system.
   * @param {object} [opts] - Options. TODO: document.
   * @throws Will throw if the path is not found.
   * @return {Promise<ReadableStream|Readable>} A `stream.Readable` in Node.js, a `ReadableStream`in the browser.
   */
  async createReadStream (path, opts = {}) {
    opts.responseType = 'stream';
    return this.readFile(path, opts)
  }

  /**
   * Get metadata about a file.
   *
   * @async
   * @param {string} path - A hyper:// URL or a relative path within the collection's file system.
   * @throws Will throw if the path is not found.
   * @return {Promise<object>} A plain object with the stat info.
   */
  async statFile (path) {
    return this.fetch(path)
  }
}

var fs = Fs;

const { SCHEMA_RESOURCE, HYPERDRIVE_SCHEME, METADATA_ID } = constants;

var resources = class Resources {
  constructor (collection) {
    this.collection = collection;
  }

  async _localDriveKey () {
    if (this._localDrive) return this._localDrive
    const drives = await this.collection.fs.listDrives();
    const writableDrives = drives.filter(f => f.writable);
    if (!writableDrives.length) throw new Error('No writable drive')
    this._localDrive = writableDrives[0].key;
    return this._localDrive
  }

  async writeFile (record, file, opts = {}) {
    const url = getContentUrl(record);
    if (!url) throw new Error('record has no file url')
    // if (!opts.metadata) opts.metadata = {}
    // opts.metadata[METADATA_ID] = record.id
    return this.collection.fs.writeFile(url, file, opts)
  }

  async readFile (record, opts = {}) {
    const url = getContentUrl(record);
    if (!url) throw new Error('record has no file url')
    return this.collection.fs.readFile(url, opts)
  }

  // TODO: Remove
  async create (value, opts) {
    return this.prepare(value, opts)
  }

  // TODO: Remove
  async prepare (value, opts = {}) {
    const { filename, prefix } = value;
    if (!filename) throw new Error('Filename is required')
    if (filename.indexOf('/') !== -1) throw new Error('Invalid filename')
    if (opts.scoped) throw new Error('Scoped option is not supported')

    let filepath;
    if (prefix) filepath = [prefix, filename].join('/');
    else filepath = filename;

    const drivekey = await this._localDriveKey();
    const contentUrl = createHyperdriveUrl(drivekey, filepath);

    let id;
    // TODO: Check for resources also/instead?
    // This checks only the local drive.
    try {
      var existing = await this.collection.fs.statFile(contentUrl);
    } catch (err) {}

    if (existing) {
      id = existing.metadata[METADATA_ID];
      if (!id) {
        if (!opts.force) throw new Error('file exists and has no resource attached. set fore to overwrite.')
      } else {
        // TODO: Preserve fields from an old resource?
        // const oldResource = await this.get({ id: existing.metadata[METADATA_ID] })
        if (!opts.update) throw new Error(`file exists, with resource ${id}. set update to overwrite.`)
      }
    }

    id = id || opts.id || Math.random();

    const record = {
      type: SCHEMA_RESOURCE,
      id,
      value: {
        ...value,
        contentUrl,
        filename
      }
    };

    return this.collection.schema.Record(record)

    // const res = await this.collection.put(record)
    // // TODO: This should get by keyseq. Or put should just return the
    // // putted record.
    // const records = await this.collection.get({
    //   id: res.id,
    //   type: SCHEMA_RESOURCE
    // }, { waitForSync: true })
    // if (!records.length) {
    //   throw new Error('error loading created resource')
    // }
    // return records[0]
  }

  resolveFileURL (record) {
    const url = getContentUrl(record);
    return this.collection.fs.resolveURL(url)
  }
};

function createHyperdriveUrl (key, path) {
  return `${HYPERDRIVE_SCHEME}//${key}/${path}`
}

function getContentUrl (record) {
  if (!record.hasType(SCHEMA_RESOURCE)) return null
  if (!record.value.contentUrl) return null
  return record.value.contentUrl
}

const debug$1 = require$$0('sonar-client');
const { Readable, Writable, Transform } = require$$1;
const { EventEmitter } = require$$2;

const { Schema, Store } = require$$3;



function uuid () {
  return base32Encode(randomBytes(16), 'Crockford').toLowerCase()
}

class Collection extends EventEmitter {
  static uuid () {
    return uuid()
  }

  /**
   * Remote collection
   *
   * @constructor
   * @param {Workspace} workspace - Remote workspace
   * @param {string} nameOrKey - Name or key of the collection
   * @return {Collection}
   */
  constructor (client, nameOrKey) {
    super();
    this.endpoint = client.endpoint + '/collection/' + nameOrKey;
    this._client = client;
    this._info = {};
    this._nameOrKey = nameOrKey;
    this._eventStreams = new Set();

    this.fs = new fs(this);
    this.resources = new resources(this);
    this.log = client.log.child({ collection: this });
    this.setMaxListeners(256);
  }

  get name () {
    if (this._info) return this._info.name
    return this._nameOrKey
  }

  get key () {
    return this._info && this._info.key
  }

  get localKey () {
    return this._info && this._info.localKey
  }

  get info () {
    return this._info
  }

  get id () {
    return this._info && this._info.id
  }

  get length () {
    return this._length || this._info.length || 0
  }

  /**
   * Populate info and schemas for this collection from server.
   *
   * @async
   * @throws Will throw if this collection does not exist or cannot be accessed.
   * @return {Promise<void>}
   */
  async open (reset = false) {
    if (!this._openPromise || reset) this._openPromise = this._open();
    return this._openPromise
  }

  async _open () {
    this._info = await this.fetch('/');

    this.schema = new Schema({ defaultNamespace: this.id });
    const typeSpecs = await this.fetch('/schema');
    for (const typeSpec of Object.values(typeSpecs)) {
      this.schema.addType(typeSpec);
    }
    this.store = new Store({ schema: this.schema });
    this.opened = true;
    this.emit('open');
  }

  /**
   * Close the collection.
   *
   * Properly closes open HTTP requests.
   * @async
   */
  async close () {
    for (const stream of this._eventStreams) {
      stream.destroy();
    }
    if (this._eventSource) this._eventSource.close();
  }

  /**
   * Put a new feed to the collection.
   *
   * @async
   * @param {string} key - The hex-encoded key of the feed to add.
   * @param {object} [info] - Optional information about the feed.
   *                          TODO: Document
   */
  async putFeed (key, info = {}) {
    return this.fetch('/feed/' + key, {
      method: 'PUT',
      body: info
    })
  }

  /**
   * @deprecated see Collection.putFeed
   */
  async addFeed (key, info = {}) {
    return this.putFeed(key, info)
  }

  /**
   * Query the collection.
   *
   * Returns an array of matching records.
   *
   * Records may have a meta property that includes query-specific metadata (e.g. the score for search queries).
   *
   * @async
   * @param {string} name - The name of a supported query.
   *                        Supported queries that ship with @arsonar/core are:
   *                        records, search, relations, history and indexes.
   * @param {object} args - The arguments for the query. Depends on the query being used.
   *                        For records: `{ schema, name, id }`
   *                        For history: `{ from: timestamp, to: timestamp }`
   *                        For search: Either a "string" for a simple full-text search, or a
   *                           tantivy query object.
   *                        For indexes: `{ schema, prop, value, from, to, reverse, limit }`
   *                           (to be documented)
   *                        For relations: `{ subject, object, predicate }`
   *                           where subject and object are ids and predicate is `type#field`
   *
   * @param {object} [opts] - Optional options
   * @param {boolean} [opts.sync=false] Wait for all pending indexing operations to be finished.
   *
   * @return {Promise<Array<Record>>} A promise that resolves to an array of record objects.
   */
  async query (name, args, opts) {
    if (this._cacheid) {
      opts.cacheid = this._cacheid;
    }

    let records = await this.fetch('/query/' + name, {
      method: 'POST',
      body: args,
      params: opts
    });

    for (const key of records.keys()) {
      const record = records[key];
      try {
        records[key] = this.store.cacheRecord(record);
      } catch (err) {
        // TODO: Where should these errors go
        console.error('Error when upcasting record', err);
        console.error('Record: ', record);
        records[key] = null;
      }
    }
    records = records.filter(r => r);

    // if (this._cacheid) {
    //   return this._cache.batch(records)
    // }

    return records
  }

  /**
   * Put a new record into the collection.
   *
   * @async
   * @param {object} record - The record.
   * @param {string} record.schema - The schema of the record.
   * @param {string} [record.id] - The entity id of the record. If empoty an id will be created.
   * @param {object} record.value - Value of the record.
   * @throws Throws if the record is invalid.
   * @return {Promise<object>} An object with an `{ id }` property.
   */
  async put (record) {
    // If the record has no id set (= is a new record), generate and add a random id.
    if (!record.id) record.id = uuid();
    // This checks if the record has type, id, value set and if the type is present
    // in the collection's schema. Throws an error if not.
    // TODO: Add feature to @arsonar/common to validate the record's value against the schema.
    record = this.schema.RecordVersion(record);
    const resultRecord = await this.fetch('/db', {
      method: 'PUT',
      body: record
    });
    return this.store.cacheRecord(resultRecord)
  }

  /**
   * Get records by their semantic address (type and id) or by their storage address (key and seq).
   *
   * @async
   * @param {object} req - The get request. Either `{ type, id }` or `{ key, seq }`.
   * @param {object} [opts] - Optional options.
   * @param {boolean} [opts.sync=false] Wait for all pending indexing operations to be finished.
   * @return {Promise<Array<object>>} A promise that resolves to an array of record objects.
   */
  async get (req, opts) {
    // TODO: Implement RecordCache.has
    // if (this._cache.has(req)) {
    //   return this._cache.get(req)
    // }
    return this.query('records', req, opts)
  }

  /**
   * Deletes a record.
   *
   * @async
   * @param {object} record - The record to delete. Has to have `{ id, type }` properties set.
   * @return {Promise<object>} - An object with `{ id, type }` properties of the deleted record.
   */
  async del (record) {
    return this.fetch('/db/' + record.id, {
      method: 'DELETE',
      params: { type: record.type }
    })
  }

  /**
   * Add a new type to the collection.
   *
   * @async
   * @param {object} schema - A schema object.
   * @throws Throws if the schema object is invalid or cannot be saved.
   * @return {Promise<object>} A promise that resolves to the saved schema object.
   */
  async putType (spec) {
    const type = this.schema.addType(spec);
    return this.fetch('/schema', {
      method: 'POST',
      body: type.toJSON()
    })
  }

  /**
   * Create a writable stream to put records into the collection.
   *
   * Example:
   * ```javascript
   * const batchStream = collection.createBatchStream()
   * batch.write(record)
   * batch.close()
   * ```
   *
   * @return {Writable<Record>} A writable stream
   */
  createBatchStream () {
    const self = this;
    const stream = new Transform({
      open (cb) {
        stream.finished = self.fetch('/db', {
          method: 'PUT',
          requestType: 'stream',
          headers: {
            'content-type': 'application/x-ndjson'
          },
          params: { batch: true },
          body: this
        }).catch(err => stream.destroy(err));
        cb();
      },
      transform (record, cb) {
        if (!record.id) record.id = uuid();
        record = self.store.cacheRecord(record);
        const json = JSON.stringify(record);
        this.push(json + '\n');
        cb();
      }
    });
    return stream
  }

  /**
   * Wait for all pending indexing operations to be finished.
   *
   * @async
   * @return {Promise<void>}
   */
  async sync () {
    return this.fetch('/sync')
  }

  /**
   * Subscribe to events on this collection.
   *
   * Returns a readable stream that emits event objects.
   * They look like this:
   * `{ event: string, data: object }`
   *
   * Events are:
   *
   * * `update`: with data `{ lseq }`
   * * `feed`: with data `{ key }`
   * * `schema-update`
   *
   * @return {Readable<object>}
   */
  createEventStream () {
    if (!this._eventStream) {
      this._initEventSource();
    }
    return this._eventStream.subscribe()
  }

  async _initEventSource () {
    if (this._eventStream) return
    this._eventStream = new TeeStream();

    const onerror = err => {
      // TODO: Where should the error go?
      // this.log.error('Error initializing event source: ' + err.message)
      // if (this._eventStream) this._eventStream.destroy(err)
      // if (this._eventStream) this._eventStream.close()
      this._eventStream = null;
    };

    try {
      if (!this._client.opened) await this._client.open();
    } catch (err) {
      onerror();
      return
    }

    this._eventSource = this._client.createEventSource('/events', {
      endpoint: this.endpoint,
      onerror,
      onmessage: (eventObject) => {
        this.log.trace('event: ' + eventObject.event);
        this._eventStream.write(eventObject);
        this.emit(eventObject.event, eventObject.data);
      }
    });
  }

  /**
   * Pull live updates from the server as they happen.
   *
   * After calling this method once, all new records and record versions
   * are pulled from the server once available. The `update` event
   * is emitted when new records are about to arrive.
   */
  pullLiveUpdates () {
    if (this._liveUpdates) return
    this._liveUpdates = true;
    const eventStream = this.createEventStream();
    eventStream.on('data', event => {
      if (event.event !== 'update') return
      const lseq = event.data.lseq;
      if (!lseq || lseq < this.length) return
      const oldLength = this.length;
      this._length = lseq + 1;
      for (let i = oldLength + 1; i < this.length; i++) {
        this.get({ lseq: i }).catch(e => {});
      }
      this.emit('update', lseq);
    });
  }

  /**
   * Subscribe to this collection.
   *
   * This will fetch all records from the first to the last and then waits for new records.
   * Currently only intended for usage in bots (not in short-running Browser clients).
   *
   * @todo: Prefix client ID to subscription name.
   * @todo: Allow to subscribe from now instead of from the beginning.
   *
   * @param {string} Subscription name. Has to be unique per running Sonar instance
   * @param {function} Async callback function that will be called for each incoming record
   */
  async subscribe (name, onRecord) {
    const self = this;
    await this._initEventSource();

    run().catch(err => {
      // TODO: How to deal with these errors?
      this.log.error('subscription error', err);
    });

    return true

    async function run () {
      // const eventStream = self.createEventStream()
      while (true) {
        // wait for one incoming update event
        const nextUpdatePromise = new Promise((resolve, reject) => {
          self.once('update', resolve);
        });

        const batch = await self._pullSubscription(name);
        for (const message of batch.messages) {
          try {
            const record = self.store.cacheRecord(message);
            await onRecord(record);
            // TODO: Do we want to ack for each message or for each batch?
            // Likely let the subscriber decide.
            await self._ackSubscription(name, message.lseq);
          } catch (err) {
            // TODO: What to do with errors here?
            self.log.error(`Error in subscription ${name}: ${err.message}`);
            debug$1(err);
            // TODO: This will easily cause an endless loop becaues it'll
            // just keep refreshing and running into errors ;-)
          }
        }
        if (batch.finished) {
          await nextUpdatePromise;
        }
      }
    }
  }

  async _pullSubscription (name, opts) {
    // TODO: Prefix name with client id.
    return this.fetch('/subscription/' + name, {
      query: opts
    })
  }

  async _ackSubscription (name, cursor) {
    // TODO: Prefix name with client id.
    return this.fetch('/subscription/' + name + '/' + cursor, {
      method: 'post'
    })
  }

  /**
   * Reindex the secondary indexes (views) in this collection.
   *
   * Use with care, this can be expensive.
   *
   * @param {Array<string>} Optional array of view names to reindex.
   *  If unset all views will be reindexed.
   */
  async reindex (views) {
    const params = {};
    if (views) params.views = views;
    return this.fetch('/reindex', {
      method: 'post',
      params
    })
  }

  async fetch (path, opts = {}) {
    if (!opts.endpoint) opts.endpoint = this.endpoint;
    return this._client.fetch(path, opts)
  }
}

class TeeStream extends Writable {
  constructor (opts) {
    super(opts);
    this._streams = new Set();
  }

  subscribe (opts) {
    const stream = new Readable(opts);
    stream.on('destroy', () => this._streams.remove(stream));
    this._streams.add(stream);
    return stream
  }

  _write (data, cb) {
    for (const stream of this._streams) {
      stream.push(data);
    }
    cb();
  }

  _destroy (err) {
    for (const stream of this._streams) {
      stream.destroy(err);
    }
    super.destroy(err);
  }
}

var collection = Collection;

const debug = require$$0('sonar:bots');

const SESSION_HEADER = 'X-Sonar-Bot-Session-Id';

class Bots {
  constructor (client) {
    this.client = client;
    this.bots = new Map();
    this.log = this.client.log.child({ name: 'bots' });
  }

  async close () {
    this._eventSource.close();
  }

  async join (botName, collection) {
    const res = await this.fetch('/join', {
      method: 'POST',
      body: { bot: botName, collection }
    });
    return res
  }

  async leave (botName, collection) {
    const res = await this.fetch('/leave', {
      method: 'POST',
      body: { bot: botName, collection }
    });
    return res
  }

  async info () {
    const res = await this.fetch('/info');
    return res
  }

  async command (botName, command, args, env) {
    const path = '/command';
    const res = await this.fetch(path, {
      method: 'POST',
      body: { bot: botName, command, args, env }
    });
    return res
  }

  async commandStatus (botName, requestId) {
    const path = `/status/${botName}/${requestId}`;
    const res = await this.fetch(path);
    return res
  }

  async getCommands () {
  }

  async _initListener () {
    this._init = true;

    // const path = this.client.endpoint + '/bot/events'
    // const headers = {
    //   ...this.client.getAuthHeaders(),
    //   ...this.getHeaders()
    // }

    const path = '/bot/events';
    this._eventSource = this.client.createEventSource(path, {
      headers: this.getHeaders(),
      onmessage: this._onmessage.bind(this),
      onerror: err => {
        // TODO: Where do these errors go?
        // TODO: After a couple of fails die.
        this.log.error({ err, message: 'Event source error' });
      }
    });
    // this._eventSource = new EventSource(path, { headers })
    // this._eventSource.addEventListener('message', message => {
    //   try {
    //     const event = JSON.parse(message.data)
    //     this._onmessage(event)
    //   } catch (e) {}
    // })
    // this._eventSource.addEventListener('error', err => {
    //   // TODO: Where do these errors go?
    //   // TODO: After a couple of fails die.
    //   this.log.error({ err, message: 'Event source error' })
    // })
  }

  async reply ({ requestId, error, result }) {
    await this.fetch('/reply', {
      method: 'POST',
      body: {
        requestId,
        error,
        result
      }
    });
  }

  async _onmessage (message) {
    try {
      const { bot: name, op, data, requestId } = message;
      const bot = this.bots.get(name);
      if (!bot) throw new Error('Unknown bot: ' + bot)
      try {
        if (op === 'join') {
          const { collection: collectionKey } = data;
          const collection = await this.client.openCollection(collectionKey);
          await bot.onjoin(collection, data.config);
          await this.reply({ requestId });
        } else if (op === 'leave') {
          const { collection: collectionKey } = data;
          const collection = await this.client.openCollection(collectionKey);
          await bot.onleave(collection);
          await this.reply({ requestId });
        } else if (op === 'command') {
          const { command, args, env } = data;
          if (env.collection) {
            env.collection = await this.client.openCollection(env.collection);
          }
          const result = await bot.oncommand(command, args, env);
          await this.reply({ requestId, result });
        }
      } catch (err) {
        this.log.error({ message: 'bot onmessage handle error: ' + err.message + ' from ' + JSON.stringify(message), err });
        debug(err);
        await this.reply({
          requestId: requestId,
          error: err.message
        });
      }
    } catch (err) {
      // TODO: Where to these errors go?
      this.log.error({ message: 'bot onmessage error', err });
    }
  }

  getHeaders () {
    const headers = {};
    if (this._sessionId) {
      headers[SESSION_HEADER] = this._sessionId;
    }
    return headers
  }

  async fetch (url, opts = {}) {
    url = '/bot' + url;
    opts.headers = opts.headers || {};
    opts.headers = { ...this.getHeaders(), ...opts.headers };
    return this.client.fetch(url, opts)
  }

  async register (name, spec, handlers) {
    if (typeof spec !== 'object') throw new Error('Spec must be an object')
    spec.name = name;
    const { config, sessionId } = await this.fetch('/register', {
      method: 'POST',
      body: spec
    });
    this._sessionId = sessionId;
    if (!this._init) await this._initListener();
    this.bots.set(name, new Bot({
      spec,
      client: this.client,
      config,
      handlers
    }));
  }

  // async configure (name, config) {
  //   const config = await this.client.fetch('/bot/configure/' + name, config)
  // }
}

class Bot {
  constructor ({ spec, config, handlers, client }) {
    this.spec = spec;
    this.name = spec.name;
    this.client = client;
    this.config = config;
    if (typeof handlers === 'function') {
      handlers = handlers({ spec, config, client });
    }
    this.handlers = handlers;
    this.sessions = new Map();
    this.opened = false;
    this.log = this.client.log.child({ name: 'bot:' + this.name });
  }

  async open () {
    if (this.opened) return
    if (this.opening) return this.opening
    let _resolve;
    this.opening = new Promise(resolve => (_resolve = resolve));
    if (this.handlers.open) await this.handlers.open();
    _resolve();
    this.opened = true;
    this.log.debug('open');
  }

  async onjoin (collection, config) {
    await this._ensureTypes(collection);
    await this.open();
    if (!this.opened && this.handlers.open) {
      await this.handlers.open();
      this.opened = true;
    }
    const session = await this.handlers.onjoin(collection, config);
    if (session.open) {
      await session.open();
    }
    if (session.onrecord) {
      collection.subscribe('bot:' + this.name, session.onrecord.bind(session));
    }
    this.sessions.set(collection.key, session);
    this.log.debug({ message: 'join', collection });
  }

  async _ensureTypes (collection) {
    if (!this.spec.types) return
    for (const typeSpec of this.spec.types) {
      if (!collection.schema.hasType(typeSpec)) {
        await collection.putType(typeSpec);
      }
    }
  }

  async onleave (collection) {
    const session = this.sessions.get(collection.key);
    if (!session) return
    if (session.close) await session.close();
    this.sessions.delete(collection.key);
    this.log.debug({ message: 'leave', collection });
  }

  async oncommand (command, args, env) {
    await this.open();

    if (!env.collection && this.handlers.oncommand) {
      this.log.debug('workspace command: ' + command);
      return await this.handlers.oncommand(command, args)
    }

    this.log.debug({ message: 'collection command: ' + command, collection: env.collection });
    const session = this.sessions.get(env.collection.key);
    if (!session) throw new Error('Bot did not join collection')
    if (!session.oncommand) throw new Error('Bot cannot handle commands')

    return await session.oncommand(command, args)
  }
}

var bots = Bots;
var Bots_1 = Bots;
bots.Bots = Bots_1;

function tryStringify (o) {
  try { return JSON.stringify(o) } catch(e) { return '"[Circular]"' }
}

var quickFormatUnescaped = format;

function format(f, args, opts) {
  var ss = (opts && opts.stringify) || tryStringify;
  var offset = 1;
  if (typeof f === 'object' && f !== null) {
    var len = args.length + offset;
    if (len === 1) return f
    var objects = new Array(len);
    objects[0] = ss(f);
    for (var index = 1; index < len; index++) {
      objects[index] = ss(args[index]);
    }
    return objects.join(' ')
  }
  if (typeof f !== 'string') {
    return f
  }
  var argLen = args.length;
  if (argLen === 0) return f
  var str = '';
  var a = 1 - offset;
  var lastPos = -1;
  var flen = (f && f.length) || 0;
  for (var i = 0; i < flen;) {
    if (f.charCodeAt(i) === 37 && i + 1 < flen) {
      lastPos = lastPos > -1 ? lastPos : 0;
      switch (f.charCodeAt(i + 1)) {
        case 100: // 'd'
          if (a >= argLen)
            break
          if (lastPos < i)
            str += f.slice(lastPos, i);
          if (args[a] == null)  break
          str += Number(args[a]);
          lastPos = i = i + 2;
          break
        case 79: // 'O'
        case 111: // 'o'
        case 106: // 'j'
          if (a >= argLen)
            break
          if (lastPos < i)
            str += f.slice(lastPos, i);
          if (args[a] === undefined) break
          var type = typeof args[a];
          if (type === 'string') {
            str += '\'' + args[a] + '\'';
            lastPos = i + 2;
            i++;
            break
          }
          if (type === 'function') {
            str += args[a].name || '<anonymous>';
            lastPos = i + 2;
            i++;
            break
          }
          str += ss(args[a]);
          lastPos = i + 2;
          i++;
          break
        case 115: // 's'
          if (a >= argLen)
            break
          if (lastPos < i)
            str += f.slice(lastPos, i);
          str += String(args[a]);
          lastPos = i + 2;
          i++;
          break
        case 37: // '%'
          if (lastPos < i)
            str += f.slice(lastPos, i);
          str += '%';
          lastPos = i + 2;
          i++;
          break
      }
      ++a;
    }
    ++i;
  }
  if (lastPos === -1)
    return f
  else if (lastPos < flen) {
    str += f.slice(lastPos);
  }

  return str
}

var browser$1 = pino;

const _console = pfGlobalThisOrFallback().console || {};
const stdSerializers = {
  mapHttpRequest: mock,
  mapHttpResponse: mock,
  wrapRequestSerializer: passthrough,
  wrapResponseSerializer: passthrough,
  wrapErrorSerializer: passthrough,
  req: mock,
  res: mock,
  err: asErrValue
};

function shouldSerialize (serialize, serializers) {
  if (Array.isArray(serialize)) {
    const hasToFilter = serialize.filter(function (k) {
      return k !== '!stdSerializers.err'
    });
    return hasToFilter
  } else if (serialize === true) {
    return Object.keys(serializers)
  }

  return false
}

function pino (opts) {
  opts = opts || {};
  opts.browser = opts.browser || {};

  const transmit = opts.browser.transmit;
  if (transmit && typeof transmit.send !== 'function') { throw Error('pino: transmit option must have a send function') }

  const proto = opts.browser.write || _console;
  if (opts.browser.write) opts.browser.asObject = true;
  const serializers = opts.serializers || {};
  const serialize = shouldSerialize(opts.browser.serialize, serializers);
  let stdErrSerialize = opts.browser.serialize;

  if (
    Array.isArray(opts.browser.serialize) &&
    opts.browser.serialize.indexOf('!stdSerializers.err') > -1
  ) stdErrSerialize = false;

  const levels = ['error', 'fatal', 'warn', 'info', 'debug', 'trace'];

  if (typeof proto === 'function') {
    proto.error = proto.fatal = proto.warn =
    proto.info = proto.debug = proto.trace = proto;
  }
  if (opts.enabled === false) opts.level = 'silent';
  const level = opts.level || 'info';
  const logger = Object.create(proto);
  if (!logger.log) logger.log = noop;

  Object.defineProperty(logger, 'levelVal', {
    get: getLevelVal
  });
  Object.defineProperty(logger, 'level', {
    get: getLevel,
    set: setLevel
  });

  const setOpts = {
    transmit,
    serialize,
    asObject: opts.browser.asObject,
    levels,
    timestamp: getTimeFunction(opts)
  };
  logger.levels = pino.levels;
  logger.level = level;

  logger.setMaxListeners = logger.getMaxListeners =
  logger.emit = logger.addListener = logger.on =
  logger.prependListener = logger.once =
  logger.prependOnceListener = logger.removeListener =
  logger.removeAllListeners = logger.listeners =
  logger.listenerCount = logger.eventNames =
  logger.write = logger.flush = noop;
  logger.serializers = serializers;
  logger._serialize = serialize;
  logger._stdErrSerialize = stdErrSerialize;
  logger.child = child;

  if (transmit) logger._logEvent = createLogEventShape();

  function getLevelVal () {
    return this.level === 'silent'
      ? Infinity
      : this.levels.values[this.level]
  }

  function getLevel () {
    return this._level
  }
  function setLevel (level) {
    if (level !== 'silent' && !this.levels.values[level]) {
      throw Error('unknown level ' + level)
    }
    this._level = level;

    set(setOpts, logger, 'error', 'log'); // <-- must stay first
    set(setOpts, logger, 'fatal', 'error');
    set(setOpts, logger, 'warn', 'error');
    set(setOpts, logger, 'info', 'log');
    set(setOpts, logger, 'debug', 'log');
    set(setOpts, logger, 'trace', 'log');
  }

  function child (bindings) {
    if (!bindings) {
      throw new Error('missing bindings for child Pino')
    }
    const bindingsSerializers = bindings.serializers;
    if (serialize && bindingsSerializers) {
      var childSerializers = Object.assign({}, serializers, bindingsSerializers);
      var childSerialize = opts.browser.serialize === true
        ? Object.keys(childSerializers)
        : serialize;
      delete bindings.serializers;
      applySerializers([bindings], childSerialize, childSerializers, this._stdErrSerialize);
    }
    function Child (parent) {
      this._childLevel = (parent._childLevel | 0) + 1;
      this.error = bind(parent, bindings, 'error');
      this.fatal = bind(parent, bindings, 'fatal');
      this.warn = bind(parent, bindings, 'warn');
      this.info = bind(parent, bindings, 'info');
      this.debug = bind(parent, bindings, 'debug');
      this.trace = bind(parent, bindings, 'trace');
      if (childSerializers) {
        this.serializers = childSerializers;
        this._serialize = childSerialize;
      }
      if (transmit) {
        this._logEvent = createLogEventShape(
          [].concat(parent._logEvent.bindings, bindings)
        );
      }
    }
    Child.prototype = this;
    return new Child(this)
  }
  return logger
}

pino.levels = {
  values: {
    fatal: 60,
    error: 50,
    warn: 40,
    info: 30,
    debug: 20,
    trace: 10
  },
  labels: {
    10: 'trace',
    20: 'debug',
    30: 'info',
    40: 'warn',
    50: 'error',
    60: 'fatal'
  }
};

pino.stdSerializers = stdSerializers;
pino.stdTimeFunctions = Object.assign({}, { nullTime, epochTime, unixTime, isoTime });

function set (opts, logger, level, fallback) {
  const proto = Object.getPrototypeOf(logger);
  logger[level] = logger.levelVal > logger.levels.values[level]
    ? noop
    : (proto[level] ? proto[level] : (_console[level] || _console[fallback] || noop));

  wrap(opts, logger, level);
}

function wrap (opts, logger, level) {
  if (!opts.transmit && logger[level] === noop) return

  logger[level] = (function (write) {
    return function LOG () {
      const ts = opts.timestamp();
      const args = new Array(arguments.length);
      const proto = (Object.getPrototypeOf && Object.getPrototypeOf(this) === _console) ? _console : this;
      for (var i = 0; i < args.length; i++) args[i] = arguments[i];

      if (opts.serialize && !opts.asObject) {
        applySerializers(args, this._serialize, this.serializers, this._stdErrSerialize);
      }
      if (opts.asObject) write.call(proto, asObject(this, level, args, ts));
      else write.apply(proto, args);

      if (opts.transmit) {
        const transmitLevel = opts.transmit.level || logger.level;
        const transmitValue = pino.levels.values[transmitLevel];
        const methodValue = pino.levels.values[level];
        if (methodValue < transmitValue) return
        transmit(this, {
          ts,
          methodLevel: level,
          methodValue,
          transmitLevel,
          transmitValue: pino.levels.values[opts.transmit.level || logger.level],
          send: opts.transmit.send,
          val: logger.levelVal
        }, args);
      }
    }
  })(logger[level]);
}

function asObject (logger, level, args, ts) {
  if (logger._serialize) applySerializers(args, logger._serialize, logger.serializers, logger._stdErrSerialize);
  const argsCloned = args.slice();
  let msg = argsCloned[0];
  const o = {};
  if (ts) {
    o.time = ts;
  }
  o.level = pino.levels.values[level];
  let lvl = (logger._childLevel | 0) + 1;
  if (lvl < 1) lvl = 1;
  // deliberate, catching objects, arrays
  if (msg !== null && typeof msg === 'object') {
    while (lvl-- && typeof argsCloned[0] === 'object') {
      Object.assign(o, argsCloned.shift());
    }
    msg = argsCloned.length ? quickFormatUnescaped(argsCloned.shift(), argsCloned) : undefined;
  } else if (typeof msg === 'string') msg = quickFormatUnescaped(argsCloned.shift(), argsCloned);
  if (msg !== undefined) o.msg = msg;
  return o
}

function applySerializers (args, serialize, serializers, stdErrSerialize) {
  for (const i in args) {
    if (stdErrSerialize && args[i] instanceof Error) {
      args[i] = pino.stdSerializers.err(args[i]);
    } else if (typeof args[i] === 'object' && !Array.isArray(args[i])) {
      for (const k in args[i]) {
        if (serialize && serialize.indexOf(k) > -1 && k in serializers) {
          args[i][k] = serializers[k](args[i][k]);
        }
      }
    }
  }
}

function bind (parent, bindings, level) {
  return function () {
    const args = new Array(1 + arguments.length);
    args[0] = bindings;
    for (var i = 1; i < args.length; i++) {
      args[i] = arguments[i - 1];
    }
    return parent[level].apply(this, args)
  }
}

function transmit (logger, opts, args) {
  const send = opts.send;
  const ts = opts.ts;
  const methodLevel = opts.methodLevel;
  const methodValue = opts.methodValue;
  const val = opts.val;
  const bindings = logger._logEvent.bindings;

  applySerializers(
    args,
    logger._serialize || Object.keys(logger.serializers),
    logger.serializers,
    logger._stdErrSerialize === undefined ? true : logger._stdErrSerialize
  );
  logger._logEvent.ts = ts;
  logger._logEvent.messages = args.filter(function (arg) {
    // bindings can only be objects, so reference equality check via indexOf is fine
    return bindings.indexOf(arg) === -1
  });

  logger._logEvent.level.label = methodLevel;
  logger._logEvent.level.value = methodValue;

  send(methodLevel, logger._logEvent, val);

  logger._logEvent = createLogEventShape(bindings);
}

function createLogEventShape (bindings) {
  return {
    ts: 0,
    messages: [],
    bindings: bindings || [],
    level: { label: '', value: 0 }
  }
}

function asErrValue (err) {
  const obj = {
    type: err.constructor.name,
    msg: err.message,
    stack: err.stack
  };
  for (const key in err) {
    if (obj[key] === undefined) {
      obj[key] = err[key];
    }
  }
  return obj
}

function getTimeFunction (opts) {
  if (typeof opts.timestamp === 'function') {
    return opts.timestamp
  }
  if (opts.timestamp === false) {
    return nullTime
  }
  return epochTime
}

function mock () { return {} }
function passthrough (a) { return a }
function noop () {}

function nullTime () { return false }
function epochTime () { return Date.now() }
function unixTime () { return Math.round(Date.now() / 1000.0) }
function isoTime () { return new Date(Date.now()).toISOString() } // using Date.now() for testability

/* eslint-disable */
/* istanbul ignore next */
function pfGlobalThisOrFallback () {
  function defd (o) { return typeof o !== 'undefined' && o }
  try {
    if (typeof globalThis !== 'undefined') return globalThis
    Object.defineProperty(Object.prototype, 'globalThis', {
      get: function () {
        delete Object.prototype.globalThis;
        return (this.globalThis = this)
      },
      configurable: true
    });
    return globalThis
  } catch (e) {
    return defd(self) || defd(window) || defd(this) || {}
  }
}

function createCommonjsModule(fn) {
  var module = { exports: {} };
	return fn(module, module.exports), module.exports;
}

var colorName = {
	"aliceblue": [240, 248, 255],
	"antiquewhite": [250, 235, 215],
	"aqua": [0, 255, 255],
	"aquamarine": [127, 255, 212],
	"azure": [240, 255, 255],
	"beige": [245, 245, 220],
	"bisque": [255, 228, 196],
	"black": [0, 0, 0],
	"blanchedalmond": [255, 235, 205],
	"blue": [0, 0, 255],
	"blueviolet": [138, 43, 226],
	"brown": [165, 42, 42],
	"burlywood": [222, 184, 135],
	"cadetblue": [95, 158, 160],
	"chartreuse": [127, 255, 0],
	"chocolate": [210, 105, 30],
	"coral": [255, 127, 80],
	"cornflowerblue": [100, 149, 237],
	"cornsilk": [255, 248, 220],
	"crimson": [220, 20, 60],
	"cyan": [0, 255, 255],
	"darkblue": [0, 0, 139],
	"darkcyan": [0, 139, 139],
	"darkgoldenrod": [184, 134, 11],
	"darkgray": [169, 169, 169],
	"darkgreen": [0, 100, 0],
	"darkgrey": [169, 169, 169],
	"darkkhaki": [189, 183, 107],
	"darkmagenta": [139, 0, 139],
	"darkolivegreen": [85, 107, 47],
	"darkorange": [255, 140, 0],
	"darkorchid": [153, 50, 204],
	"darkred": [139, 0, 0],
	"darksalmon": [233, 150, 122],
	"darkseagreen": [143, 188, 143],
	"darkslateblue": [72, 61, 139],
	"darkslategray": [47, 79, 79],
	"darkslategrey": [47, 79, 79],
	"darkturquoise": [0, 206, 209],
	"darkviolet": [148, 0, 211],
	"deeppink": [255, 20, 147],
	"deepskyblue": [0, 191, 255],
	"dimgray": [105, 105, 105],
	"dimgrey": [105, 105, 105],
	"dodgerblue": [30, 144, 255],
	"firebrick": [178, 34, 34],
	"floralwhite": [255, 250, 240],
	"forestgreen": [34, 139, 34],
	"fuchsia": [255, 0, 255],
	"gainsboro": [220, 220, 220],
	"ghostwhite": [248, 248, 255],
	"gold": [255, 215, 0],
	"goldenrod": [218, 165, 32],
	"gray": [128, 128, 128],
	"green": [0, 128, 0],
	"greenyellow": [173, 255, 47],
	"grey": [128, 128, 128],
	"honeydew": [240, 255, 240],
	"hotpink": [255, 105, 180],
	"indianred": [205, 92, 92],
	"indigo": [75, 0, 130],
	"ivory": [255, 255, 240],
	"khaki": [240, 230, 140],
	"lavender": [230, 230, 250],
	"lavenderblush": [255, 240, 245],
	"lawngreen": [124, 252, 0],
	"lemonchiffon": [255, 250, 205],
	"lightblue": [173, 216, 230],
	"lightcoral": [240, 128, 128],
	"lightcyan": [224, 255, 255],
	"lightgoldenrodyellow": [250, 250, 210],
	"lightgray": [211, 211, 211],
	"lightgreen": [144, 238, 144],
	"lightgrey": [211, 211, 211],
	"lightpink": [255, 182, 193],
	"lightsalmon": [255, 160, 122],
	"lightseagreen": [32, 178, 170],
	"lightskyblue": [135, 206, 250],
	"lightslategray": [119, 136, 153],
	"lightslategrey": [119, 136, 153],
	"lightsteelblue": [176, 196, 222],
	"lightyellow": [255, 255, 224],
	"lime": [0, 255, 0],
	"limegreen": [50, 205, 50],
	"linen": [250, 240, 230],
	"magenta": [255, 0, 255],
	"maroon": [128, 0, 0],
	"mediumaquamarine": [102, 205, 170],
	"mediumblue": [0, 0, 205],
	"mediumorchid": [186, 85, 211],
	"mediumpurple": [147, 112, 219],
	"mediumseagreen": [60, 179, 113],
	"mediumslateblue": [123, 104, 238],
	"mediumspringgreen": [0, 250, 154],
	"mediumturquoise": [72, 209, 204],
	"mediumvioletred": [199, 21, 133],
	"midnightblue": [25, 25, 112],
	"mintcream": [245, 255, 250],
	"mistyrose": [255, 228, 225],
	"moccasin": [255, 228, 181],
	"navajowhite": [255, 222, 173],
	"navy": [0, 0, 128],
	"oldlace": [253, 245, 230],
	"olive": [128, 128, 0],
	"olivedrab": [107, 142, 35],
	"orange": [255, 165, 0],
	"orangered": [255, 69, 0],
	"orchid": [218, 112, 214],
	"palegoldenrod": [238, 232, 170],
	"palegreen": [152, 251, 152],
	"paleturquoise": [175, 238, 238],
	"palevioletred": [219, 112, 147],
	"papayawhip": [255, 239, 213],
	"peachpuff": [255, 218, 185],
	"peru": [205, 133, 63],
	"pink": [255, 192, 203],
	"plum": [221, 160, 221],
	"powderblue": [176, 224, 230],
	"purple": [128, 0, 128],
	"rebeccapurple": [102, 51, 153],
	"red": [255, 0, 0],
	"rosybrown": [188, 143, 143],
	"royalblue": [65, 105, 225],
	"saddlebrown": [139, 69, 19],
	"salmon": [250, 128, 114],
	"sandybrown": [244, 164, 96],
	"seagreen": [46, 139, 87],
	"seashell": [255, 245, 238],
	"sienna": [160, 82, 45],
	"silver": [192, 192, 192],
	"skyblue": [135, 206, 235],
	"slateblue": [106, 90, 205],
	"slategray": [112, 128, 144],
	"slategrey": [112, 128, 144],
	"snow": [255, 250, 250],
	"springgreen": [0, 255, 127],
	"steelblue": [70, 130, 180],
	"tan": [210, 180, 140],
	"teal": [0, 128, 128],
	"thistle": [216, 191, 216],
	"tomato": [255, 99, 71],
	"turquoise": [64, 224, 208],
	"violet": [238, 130, 238],
	"wheat": [245, 222, 179],
	"white": [255, 255, 255],
	"whitesmoke": [245, 245, 245],
	"yellow": [255, 255, 0],
	"yellowgreen": [154, 205, 50]
};

/* MIT license */

/* eslint-disable no-mixed-operators */


// NOTE: conversions should only return primitive values (i.e. arrays, or
//       values that give correct `typeof` results).
//       do not use box values types (i.e. Number(), String(), etc.)

const reverseKeywords = {};
for (const key of Object.keys(colorName)) {
	reverseKeywords[colorName[key]] = key;
}

const convert$1 = {
	rgb: {channels: 3, labels: 'rgb'},
	hsl: {channels: 3, labels: 'hsl'},
	hsv: {channels: 3, labels: 'hsv'},
	hwb: {channels: 3, labels: 'hwb'},
	cmyk: {channels: 4, labels: 'cmyk'},
	xyz: {channels: 3, labels: 'xyz'},
	lab: {channels: 3, labels: 'lab'},
	lch: {channels: 3, labels: 'lch'},
	hex: {channels: 1, labels: ['hex']},
	keyword: {channels: 1, labels: ['keyword']},
	ansi16: {channels: 1, labels: ['ansi16']},
	ansi256: {channels: 1, labels: ['ansi256']},
	hcg: {channels: 3, labels: ['h', 'c', 'g']},
	apple: {channels: 3, labels: ['r16', 'g16', 'b16']},
	gray: {channels: 1, labels: ['gray']}
};

var conversions = convert$1;

// Hide .channels and .labels properties
for (const model of Object.keys(convert$1)) {
	if (!('channels' in convert$1[model])) {
		throw new Error('missing channels property: ' + model);
	}

	if (!('labels' in convert$1[model])) {
		throw new Error('missing channel labels property: ' + model);
	}

	if (convert$1[model].labels.length !== convert$1[model].channels) {
		throw new Error('channel and label counts mismatch: ' + model);
	}

	const {channels, labels} = convert$1[model];
	delete convert$1[model].channels;
	delete convert$1[model].labels;
	Object.defineProperty(convert$1[model], 'channels', {value: channels});
	Object.defineProperty(convert$1[model], 'labels', {value: labels});
}

convert$1.rgb.hsl = function (rgb) {
	const r = rgb[0] / 255;
	const g = rgb[1] / 255;
	const b = rgb[2] / 255;
	const min = Math.min(r, g, b);
	const max = Math.max(r, g, b);
	const delta = max - min;
	let h;
	let s;

	if (max === min) {
		h = 0;
	} else if (r === max) {
		h = (g - b) / delta;
	} else if (g === max) {
		h = 2 + (b - r) / delta;
	} else if (b === max) {
		h = 4 + (r - g) / delta;
	}

	h = Math.min(h * 60, 360);

	if (h < 0) {
		h += 360;
	}

	const l = (min + max) / 2;

	if (max === min) {
		s = 0;
	} else if (l <= 0.5) {
		s = delta / (max + min);
	} else {
		s = delta / (2 - max - min);
	}

	return [h, s * 100, l * 100];
};

convert$1.rgb.hsv = function (rgb) {
	let rdif;
	let gdif;
	let bdif;
	let h;
	let s;

	const r = rgb[0] / 255;
	const g = rgb[1] / 255;
	const b = rgb[2] / 255;
	const v = Math.max(r, g, b);
	const diff = v - Math.min(r, g, b);
	const diffc = function (c) {
		return (v - c) / 6 / diff + 1 / 2;
	};

	if (diff === 0) {
		h = 0;
		s = 0;
	} else {
		s = diff / v;
		rdif = diffc(r);
		gdif = diffc(g);
		bdif = diffc(b);

		if (r === v) {
			h = bdif - gdif;
		} else if (g === v) {
			h = (1 / 3) + rdif - bdif;
		} else if (b === v) {
			h = (2 / 3) + gdif - rdif;
		}

		if (h < 0) {
			h += 1;
		} else if (h > 1) {
			h -= 1;
		}
	}

	return [
		h * 360,
		s * 100,
		v * 100
	];
};

convert$1.rgb.hwb = function (rgb) {
	const r = rgb[0];
	const g = rgb[1];
	let b = rgb[2];
	const h = convert$1.rgb.hsl(rgb)[0];
	const w = 1 / 255 * Math.min(r, Math.min(g, b));

	b = 1 - 1 / 255 * Math.max(r, Math.max(g, b));

	return [h, w * 100, b * 100];
};

convert$1.rgb.cmyk = function (rgb) {
	const r = rgb[0] / 255;
	const g = rgb[1] / 255;
	const b = rgb[2] / 255;

	const k = Math.min(1 - r, 1 - g, 1 - b);
	const c = (1 - r - k) / (1 - k) || 0;
	const m = (1 - g - k) / (1 - k) || 0;
	const y = (1 - b - k) / (1 - k) || 0;

	return [c * 100, m * 100, y * 100, k * 100];
};

function comparativeDistance(x, y) {
	/*
		See https://en.m.wikipedia.org/wiki/Euclidean_distance#Squared_Euclidean_distance
	*/
	return (
		((x[0] - y[0]) ** 2) +
		((x[1] - y[1]) ** 2) +
		((x[2] - y[2]) ** 2)
	);
}

convert$1.rgb.keyword = function (rgb) {
	const reversed = reverseKeywords[rgb];
	if (reversed) {
		return reversed;
	}

	let currentClosestDistance = Infinity;
	let currentClosestKeyword;

	for (const keyword of Object.keys(colorName)) {
		const value = colorName[keyword];

		// Compute comparative distance
		const distance = comparativeDistance(rgb, value);

		// Check if its less, if so set as closest
		if (distance < currentClosestDistance) {
			currentClosestDistance = distance;
			currentClosestKeyword = keyword;
		}
	}

	return currentClosestKeyword;
};

convert$1.keyword.rgb = function (keyword) {
	return colorName[keyword];
};

convert$1.rgb.xyz = function (rgb) {
	let r = rgb[0] / 255;
	let g = rgb[1] / 255;
	let b = rgb[2] / 255;

	// Assume sRGB
	r = r > 0.04045 ? (((r + 0.055) / 1.055) ** 2.4) : (r / 12.92);
	g = g > 0.04045 ? (((g + 0.055) / 1.055) ** 2.4) : (g / 12.92);
	b = b > 0.04045 ? (((b + 0.055) / 1.055) ** 2.4) : (b / 12.92);

	const x = (r * 0.4124) + (g * 0.3576) + (b * 0.1805);
	const y = (r * 0.2126) + (g * 0.7152) + (b * 0.0722);
	const z = (r * 0.0193) + (g * 0.1192) + (b * 0.9505);

	return [x * 100, y * 100, z * 100];
};

convert$1.rgb.lab = function (rgb) {
	const xyz = convert$1.rgb.xyz(rgb);
	let x = xyz[0];
	let y = xyz[1];
	let z = xyz[2];

	x /= 95.047;
	y /= 100;
	z /= 108.883;

	x = x > 0.008856 ? (x ** (1 / 3)) : (7.787 * x) + (16 / 116);
	y = y > 0.008856 ? (y ** (1 / 3)) : (7.787 * y) + (16 / 116);
	z = z > 0.008856 ? (z ** (1 / 3)) : (7.787 * z) + (16 / 116);

	const l = (116 * y) - 16;
	const a = 500 * (x - y);
	const b = 200 * (y - z);

	return [l, a, b];
};

convert$1.hsl.rgb = function (hsl) {
	const h = hsl[0] / 360;
	const s = hsl[1] / 100;
	const l = hsl[2] / 100;
	let t2;
	let t3;
	let val;

	if (s === 0) {
		val = l * 255;
		return [val, val, val];
	}

	if (l < 0.5) {
		t2 = l * (1 + s);
	} else {
		t2 = l + s - l * s;
	}

	const t1 = 2 * l - t2;

	const rgb = [0, 0, 0];
	for (let i = 0; i < 3; i++) {
		t3 = h + 1 / 3 * -(i - 1);
		if (t3 < 0) {
			t3++;
		}

		if (t3 > 1) {
			t3--;
		}

		if (6 * t3 < 1) {
			val = t1 + (t2 - t1) * 6 * t3;
		} else if (2 * t3 < 1) {
			val = t2;
		} else if (3 * t3 < 2) {
			val = t1 + (t2 - t1) * (2 / 3 - t3) * 6;
		} else {
			val = t1;
		}

		rgb[i] = val * 255;
	}

	return rgb;
};

convert$1.hsl.hsv = function (hsl) {
	const h = hsl[0];
	let s = hsl[1] / 100;
	let l = hsl[2] / 100;
	let smin = s;
	const lmin = Math.max(l, 0.01);

	l *= 2;
	s *= (l <= 1) ? l : 2 - l;
	smin *= lmin <= 1 ? lmin : 2 - lmin;
	const v = (l + s) / 2;
	const sv = l === 0 ? (2 * smin) / (lmin + smin) : (2 * s) / (l + s);

	return [h, sv * 100, v * 100];
};

convert$1.hsv.rgb = function (hsv) {
	const h = hsv[0] / 60;
	const s = hsv[1] / 100;
	let v = hsv[2] / 100;
	const hi = Math.floor(h) % 6;

	const f = h - Math.floor(h);
	const p = 255 * v * (1 - s);
	const q = 255 * v * (1 - (s * f));
	const t = 255 * v * (1 - (s * (1 - f)));
	v *= 255;

	switch (hi) {
		case 0:
			return [v, t, p];
		case 1:
			return [q, v, p];
		case 2:
			return [p, v, t];
		case 3:
			return [p, q, v];
		case 4:
			return [t, p, v];
		case 5:
			return [v, p, q];
	}
};

convert$1.hsv.hsl = function (hsv) {
	const h = hsv[0];
	const s = hsv[1] / 100;
	const v = hsv[2] / 100;
	const vmin = Math.max(v, 0.01);
	let sl;
	let l;

	l = (2 - s) * v;
	const lmin = (2 - s) * vmin;
	sl = s * vmin;
	sl /= (lmin <= 1) ? lmin : 2 - lmin;
	sl = sl || 0;
	l /= 2;

	return [h, sl * 100, l * 100];
};

// http://dev.w3.org/csswg/css-color/#hwb-to-rgb
convert$1.hwb.rgb = function (hwb) {
	const h = hwb[0] / 360;
	let wh = hwb[1] / 100;
	let bl = hwb[2] / 100;
	const ratio = wh + bl;
	let f;

	// Wh + bl cant be > 1
	if (ratio > 1) {
		wh /= ratio;
		bl /= ratio;
	}

	const i = Math.floor(6 * h);
	const v = 1 - bl;
	f = 6 * h - i;

	if ((i & 0x01) !== 0) {
		f = 1 - f;
	}

	const n = wh + f * (v - wh); // Linear interpolation

	let r;
	let g;
	let b;
	/* eslint-disable max-statements-per-line,no-multi-spaces */
	switch (i) {
		default:
		case 6:
		case 0: r = v;  g = n;  b = wh; break;
		case 1: r = n;  g = v;  b = wh; break;
		case 2: r = wh; g = v;  b = n; break;
		case 3: r = wh; g = n;  b = v; break;
		case 4: r = n;  g = wh; b = v; break;
		case 5: r = v;  g = wh; b = n; break;
	}
	/* eslint-enable max-statements-per-line,no-multi-spaces */

	return [r * 255, g * 255, b * 255];
};

convert$1.cmyk.rgb = function (cmyk) {
	const c = cmyk[0] / 100;
	const m = cmyk[1] / 100;
	const y = cmyk[2] / 100;
	const k = cmyk[3] / 100;

	const r = 1 - Math.min(1, c * (1 - k) + k);
	const g = 1 - Math.min(1, m * (1 - k) + k);
	const b = 1 - Math.min(1, y * (1 - k) + k);

	return [r * 255, g * 255, b * 255];
};

convert$1.xyz.rgb = function (xyz) {
	const x = xyz[0] / 100;
	const y = xyz[1] / 100;
	const z = xyz[2] / 100;
	let r;
	let g;
	let b;

	r = (x * 3.2406) + (y * -1.5372) + (z * -0.4986);
	g = (x * -0.9689) + (y * 1.8758) + (z * 0.0415);
	b = (x * 0.0557) + (y * -0.2040) + (z * 1.0570);

	// Assume sRGB
	r = r > 0.0031308
		? ((1.055 * (r ** (1.0 / 2.4))) - 0.055)
		: r * 12.92;

	g = g > 0.0031308
		? ((1.055 * (g ** (1.0 / 2.4))) - 0.055)
		: g * 12.92;

	b = b > 0.0031308
		? ((1.055 * (b ** (1.0 / 2.4))) - 0.055)
		: b * 12.92;

	r = Math.min(Math.max(0, r), 1);
	g = Math.min(Math.max(0, g), 1);
	b = Math.min(Math.max(0, b), 1);

	return [r * 255, g * 255, b * 255];
};

convert$1.xyz.lab = function (xyz) {
	let x = xyz[0];
	let y = xyz[1];
	let z = xyz[2];

	x /= 95.047;
	y /= 100;
	z /= 108.883;

	x = x > 0.008856 ? (x ** (1 / 3)) : (7.787 * x) + (16 / 116);
	y = y > 0.008856 ? (y ** (1 / 3)) : (7.787 * y) + (16 / 116);
	z = z > 0.008856 ? (z ** (1 / 3)) : (7.787 * z) + (16 / 116);

	const l = (116 * y) - 16;
	const a = 500 * (x - y);
	const b = 200 * (y - z);

	return [l, a, b];
};

convert$1.lab.xyz = function (lab) {
	const l = lab[0];
	const a = lab[1];
	const b = lab[2];
	let x;
	let y;
	let z;

	y = (l + 16) / 116;
	x = a / 500 + y;
	z = y - b / 200;

	const y2 = y ** 3;
	const x2 = x ** 3;
	const z2 = z ** 3;
	y = y2 > 0.008856 ? y2 : (y - 16 / 116) / 7.787;
	x = x2 > 0.008856 ? x2 : (x - 16 / 116) / 7.787;
	z = z2 > 0.008856 ? z2 : (z - 16 / 116) / 7.787;

	x *= 95.047;
	y *= 100;
	z *= 108.883;

	return [x, y, z];
};

convert$1.lab.lch = function (lab) {
	const l = lab[0];
	const a = lab[1];
	const b = lab[2];
	let h;

	const hr = Math.atan2(b, a);
	h = hr * 360 / 2 / Math.PI;

	if (h < 0) {
		h += 360;
	}

	const c = Math.sqrt(a * a + b * b);

	return [l, c, h];
};

convert$1.lch.lab = function (lch) {
	const l = lch[0];
	const c = lch[1];
	const h = lch[2];

	const hr = h / 360 * 2 * Math.PI;
	const a = c * Math.cos(hr);
	const b = c * Math.sin(hr);

	return [l, a, b];
};

convert$1.rgb.ansi16 = function (args, saturation = null) {
	const [r, g, b] = args;
	let value = saturation === null ? convert$1.rgb.hsv(args)[2] : saturation; // Hsv -> ansi16 optimization

	value = Math.round(value / 50);

	if (value === 0) {
		return 30;
	}

	let ansi = 30
		+ ((Math.round(b / 255) << 2)
		| (Math.round(g / 255) << 1)
		| Math.round(r / 255));

	if (value === 2) {
		ansi += 60;
	}

	return ansi;
};

convert$1.hsv.ansi16 = function (args) {
	// Optimization here; we already know the value and don't need to get
	// it converted for us.
	return convert$1.rgb.ansi16(convert$1.hsv.rgb(args), args[2]);
};

convert$1.rgb.ansi256 = function (args) {
	const r = args[0];
	const g = args[1];
	const b = args[2];

	// We use the extended greyscale palette here, with the exception of
	// black and white. normal palette only has 4 greyscale shades.
	if (r === g && g === b) {
		if (r < 8) {
			return 16;
		}

		if (r > 248) {
			return 231;
		}

		return Math.round(((r - 8) / 247) * 24) + 232;
	}

	const ansi = 16
		+ (36 * Math.round(r / 255 * 5))
		+ (6 * Math.round(g / 255 * 5))
		+ Math.round(b / 255 * 5);

	return ansi;
};

convert$1.ansi16.rgb = function (args) {
	let color = args % 10;

	// Handle greyscale
	if (color === 0 || color === 7) {
		if (args > 50) {
			color += 3.5;
		}

		color = color / 10.5 * 255;

		return [color, color, color];
	}

	const mult = (~~(args > 50) + 1) * 0.5;
	const r = ((color & 1) * mult) * 255;
	const g = (((color >> 1) & 1) * mult) * 255;
	const b = (((color >> 2) & 1) * mult) * 255;

	return [r, g, b];
};

convert$1.ansi256.rgb = function (args) {
	// Handle greyscale
	if (args >= 232) {
		const c = (args - 232) * 10 + 8;
		return [c, c, c];
	}

	args -= 16;

	let rem;
	const r = Math.floor(args / 36) / 5 * 255;
	const g = Math.floor((rem = args % 36) / 6) / 5 * 255;
	const b = (rem % 6) / 5 * 255;

	return [r, g, b];
};

convert$1.rgb.hex = function (args) {
	const integer = ((Math.round(args[0]) & 0xFF) << 16)
		+ ((Math.round(args[1]) & 0xFF) << 8)
		+ (Math.round(args[2]) & 0xFF);

	const string = integer.toString(16).toUpperCase();
	return '000000'.substring(string.length) + string;
};

convert$1.hex.rgb = function (args) {
	const match = args.toString(16).match(/[a-f0-9]{6}|[a-f0-9]{3}/i);
	if (!match) {
		return [0, 0, 0];
	}

	let colorString = match[0];

	if (match[0].length === 3) {
		colorString = colorString.split('').map(char => {
			return char + char;
		}).join('');
	}

	const integer = parseInt(colorString, 16);
	const r = (integer >> 16) & 0xFF;
	const g = (integer >> 8) & 0xFF;
	const b = integer & 0xFF;

	return [r, g, b];
};

convert$1.rgb.hcg = function (rgb) {
	const r = rgb[0] / 255;
	const g = rgb[1] / 255;
	const b = rgb[2] / 255;
	const max = Math.max(Math.max(r, g), b);
	const min = Math.min(Math.min(r, g), b);
	const chroma = (max - min);
	let grayscale;
	let hue;

	if (chroma < 1) {
		grayscale = min / (1 - chroma);
	} else {
		grayscale = 0;
	}

	if (chroma <= 0) {
		hue = 0;
	} else
	if (max === r) {
		hue = ((g - b) / chroma) % 6;
	} else
	if (max === g) {
		hue = 2 + (b - r) / chroma;
	} else {
		hue = 4 + (r - g) / chroma;
	}

	hue /= 6;
	hue %= 1;

	return [hue * 360, chroma * 100, grayscale * 100];
};

convert$1.hsl.hcg = function (hsl) {
	const s = hsl[1] / 100;
	const l = hsl[2] / 100;

	const c = l < 0.5 ? (2.0 * s * l) : (2.0 * s * (1.0 - l));

	let f = 0;
	if (c < 1.0) {
		f = (l - 0.5 * c) / (1.0 - c);
	}

	return [hsl[0], c * 100, f * 100];
};

convert$1.hsv.hcg = function (hsv) {
	const s = hsv[1] / 100;
	const v = hsv[2] / 100;

	const c = s * v;
	let f = 0;

	if (c < 1.0) {
		f = (v - c) / (1 - c);
	}

	return [hsv[0], c * 100, f * 100];
};

convert$1.hcg.rgb = function (hcg) {
	const h = hcg[0] / 360;
	const c = hcg[1] / 100;
	const g = hcg[2] / 100;

	if (c === 0.0) {
		return [g * 255, g * 255, g * 255];
	}

	const pure = [0, 0, 0];
	const hi = (h % 1) * 6;
	const v = hi % 1;
	const w = 1 - v;
	let mg = 0;

	/* eslint-disable max-statements-per-line */
	switch (Math.floor(hi)) {
		case 0:
			pure[0] = 1; pure[1] = v; pure[2] = 0; break;
		case 1:
			pure[0] = w; pure[1] = 1; pure[2] = 0; break;
		case 2:
			pure[0] = 0; pure[1] = 1; pure[2] = v; break;
		case 3:
			pure[0] = 0; pure[1] = w; pure[2] = 1; break;
		case 4:
			pure[0] = v; pure[1] = 0; pure[2] = 1; break;
		default:
			pure[0] = 1; pure[1] = 0; pure[2] = w;
	}
	/* eslint-enable max-statements-per-line */

	mg = (1.0 - c) * g;

	return [
		(c * pure[0] + mg) * 255,
		(c * pure[1] + mg) * 255,
		(c * pure[2] + mg) * 255
	];
};

convert$1.hcg.hsv = function (hcg) {
	const c = hcg[1] / 100;
	const g = hcg[2] / 100;

	const v = c + g * (1.0 - c);
	let f = 0;

	if (v > 0.0) {
		f = c / v;
	}

	return [hcg[0], f * 100, v * 100];
};

convert$1.hcg.hsl = function (hcg) {
	const c = hcg[1] / 100;
	const g = hcg[2] / 100;

	const l = g * (1.0 - c) + 0.5 * c;
	let s = 0;

	if (l > 0.0 && l < 0.5) {
		s = c / (2 * l);
	} else
	if (l >= 0.5 && l < 1.0) {
		s = c / (2 * (1 - l));
	}

	return [hcg[0], s * 100, l * 100];
};

convert$1.hcg.hwb = function (hcg) {
	const c = hcg[1] / 100;
	const g = hcg[2] / 100;
	const v = c + g * (1.0 - c);
	return [hcg[0], (v - c) * 100, (1 - v) * 100];
};

convert$1.hwb.hcg = function (hwb) {
	const w = hwb[1] / 100;
	const b = hwb[2] / 100;
	const v = 1 - b;
	const c = v - w;
	let g = 0;

	if (c < 1) {
		g = (v - c) / (1 - c);
	}

	return [hwb[0], c * 100, g * 100];
};

convert$1.apple.rgb = function (apple) {
	return [(apple[0] / 65535) * 255, (apple[1] / 65535) * 255, (apple[2] / 65535) * 255];
};

convert$1.rgb.apple = function (rgb) {
	return [(rgb[0] / 255) * 65535, (rgb[1] / 255) * 65535, (rgb[2] / 255) * 65535];
};

convert$1.gray.rgb = function (args) {
	return [args[0] / 100 * 255, args[0] / 100 * 255, args[0] / 100 * 255];
};

convert$1.gray.hsl = function (args) {
	return [0, 0, args[0]];
};

convert$1.gray.hsv = convert$1.gray.hsl;

convert$1.gray.hwb = function (gray) {
	return [0, 100, gray[0]];
};

convert$1.gray.cmyk = function (gray) {
	return [0, 0, 0, gray[0]];
};

convert$1.gray.lab = function (gray) {
	return [gray[0], 0, 0];
};

convert$1.gray.hex = function (gray) {
	const val = Math.round(gray[0] / 100 * 255) & 0xFF;
	const integer = (val << 16) + (val << 8) + val;

	const string = integer.toString(16).toUpperCase();
	return '000000'.substring(string.length) + string;
};

convert$1.rgb.gray = function (rgb) {
	const val = (rgb[0] + rgb[1] + rgb[2]) / 3;
	return [val / 255 * 100];
};

/*
	This function routes a model to all other models.

	all functions that are routed have a property `.conversion` attached
	to the returned synthetic function. This property is an array
	of strings, each with the steps in between the 'from' and 'to'
	color models (inclusive).

	conversions that are not possible simply are not included.
*/

function buildGraph() {
	const graph = {};
	// https://jsperf.com/object-keys-vs-for-in-with-closure/3
	const models = Object.keys(conversions);

	for (let len = models.length, i = 0; i < len; i++) {
		graph[models[i]] = {
			// http://jsperf.com/1-vs-infinity
			// micro-opt, but this is simple.
			distance: -1,
			parent: null
		};
	}

	return graph;
}

// https://en.wikipedia.org/wiki/Breadth-first_search
function deriveBFS(fromModel) {
	const graph = buildGraph();
	const queue = [fromModel]; // Unshift -> queue -> pop

	graph[fromModel].distance = 0;

	while (queue.length) {
		const current = queue.pop();
		const adjacents = Object.keys(conversions[current]);

		for (let len = adjacents.length, i = 0; i < len; i++) {
			const adjacent = adjacents[i];
			const node = graph[adjacent];

			if (node.distance === -1) {
				node.distance = graph[current].distance + 1;
				node.parent = current;
				queue.unshift(adjacent);
			}
		}
	}

	return graph;
}

function link(from, to) {
	return function (args) {
		return to(from(args));
	};
}

function wrapConversion(toModel, graph) {
	const path = [graph[toModel].parent, toModel];
	let fn = conversions[graph[toModel].parent][toModel];

	let cur = graph[toModel].parent;
	while (graph[cur].parent) {
		path.unshift(graph[cur].parent);
		fn = link(conversions[graph[cur].parent][cur], fn);
		cur = graph[cur].parent;
	}

	fn.conversion = path;
	return fn;
}

var route = function (fromModel) {
	const graph = deriveBFS(fromModel);
	const conversion = {};

	const models = Object.keys(graph);
	for (let len = models.length, i = 0; i < len; i++) {
		const toModel = models[i];
		const node = graph[toModel];

		if (node.parent === null) {
			// No possible conversion, or this node is the source model.
			continue;
		}

		conversion[toModel] = wrapConversion(toModel, graph);
	}

	return conversion;
};

const convert = {};

const models = Object.keys(conversions);

function wrapRaw(fn) {
	const wrappedFn = function (...args) {
		const arg0 = args[0];
		if (arg0 === undefined || arg0 === null) {
			return arg0;
		}

		if (arg0.length > 1) {
			args = arg0;
		}

		return fn(args);
	};

	// Preserve .conversion property if there is one
	if ('conversion' in fn) {
		wrappedFn.conversion = fn.conversion;
	}

	return wrappedFn;
}

function wrapRounded(fn) {
	const wrappedFn = function (...args) {
		const arg0 = args[0];

		if (arg0 === undefined || arg0 === null) {
			return arg0;
		}

		if (arg0.length > 1) {
			args = arg0;
		}

		const result = fn(args);

		// We're assuming the result is an array here.
		// see notice in conversions.js; don't use box types
		// in conversion functions.
		if (typeof result === 'object') {
			for (let len = result.length, i = 0; i < len; i++) {
				result[i] = Math.round(result[i]);
			}
		}

		return result;
	};

	// Preserve .conversion property if there is one
	if ('conversion' in fn) {
		wrappedFn.conversion = fn.conversion;
	}

	return wrappedFn;
}

models.forEach(fromModel => {
	convert[fromModel] = {};

	Object.defineProperty(convert[fromModel], 'channels', {value: conversions[fromModel].channels});
	Object.defineProperty(convert[fromModel], 'labels', {value: conversions[fromModel].labels});

	const routes = route(fromModel);
	const routeModels = Object.keys(routes);

	routeModels.forEach(toModel => {
		const fn = routes[toModel];

		convert[fromModel][toModel] = wrapRounded(fn);
		convert[fromModel][toModel].raw = wrapRaw(fn);
	});
});

var colorConvert = convert;

var ansiStyles = createCommonjsModule(function (module) {

const wrapAnsi16 = (fn, offset) => (...args) => {
	const code = fn(...args);
	return `\u001B[${code + offset}m`;
};

const wrapAnsi256 = (fn, offset) => (...args) => {
	const code = fn(...args);
	return `\u001B[${38 + offset};5;${code}m`;
};

const wrapAnsi16m = (fn, offset) => (...args) => {
	const rgb = fn(...args);
	return `\u001B[${38 + offset};2;${rgb[0]};${rgb[1]};${rgb[2]}m`;
};

const ansi2ansi = n => n;
const rgb2rgb = (r, g, b) => [r, g, b];

const setLazyProperty = (object, property, get) => {
	Object.defineProperty(object, property, {
		get: () => {
			const value = get();

			Object.defineProperty(object, property, {
				value,
				enumerable: true,
				configurable: true
			});

			return value;
		},
		enumerable: true,
		configurable: true
	});
};

/** @type {typeof import('color-convert')} */
let colorConvert$1;
const makeDynamicStyles = (wrap, targetSpace, identity, isBackground) => {
	if (colorConvert$1 === undefined) {
		colorConvert$1 = colorConvert;
	}

	const offset = isBackground ? 10 : 0;
	const styles = {};

	for (const [sourceSpace, suite] of Object.entries(colorConvert$1)) {
		const name = sourceSpace === 'ansi16' ? 'ansi' : sourceSpace;
		if (sourceSpace === targetSpace) {
			styles[name] = wrap(identity, offset);
		} else if (typeof suite === 'object') {
			styles[name] = wrap(suite[targetSpace], offset);
		}
	}

	return styles;
};

function assembleStyles() {
	const codes = new Map();
	const styles = {
		modifier: {
			reset: [0, 0],
			// 21 isn't widely supported and 22 does the same thing
			bold: [1, 22],
			dim: [2, 22],
			italic: [3, 23],
			underline: [4, 24],
			inverse: [7, 27],
			hidden: [8, 28],
			strikethrough: [9, 29]
		},
		color: {
			black: [30, 39],
			red: [31, 39],
			green: [32, 39],
			yellow: [33, 39],
			blue: [34, 39],
			magenta: [35, 39],
			cyan: [36, 39],
			white: [37, 39],

			// Bright color
			blackBright: [90, 39],
			redBright: [91, 39],
			greenBright: [92, 39],
			yellowBright: [93, 39],
			blueBright: [94, 39],
			magentaBright: [95, 39],
			cyanBright: [96, 39],
			whiteBright: [97, 39]
		},
		bgColor: {
			bgBlack: [40, 49],
			bgRed: [41, 49],
			bgGreen: [42, 49],
			bgYellow: [43, 49],
			bgBlue: [44, 49],
			bgMagenta: [45, 49],
			bgCyan: [46, 49],
			bgWhite: [47, 49],

			// Bright color
			bgBlackBright: [100, 49],
			bgRedBright: [101, 49],
			bgGreenBright: [102, 49],
			bgYellowBright: [103, 49],
			bgBlueBright: [104, 49],
			bgMagentaBright: [105, 49],
			bgCyanBright: [106, 49],
			bgWhiteBright: [107, 49]
		}
	};

	// Alias bright black as gray (and grey)
	styles.color.gray = styles.color.blackBright;
	styles.bgColor.bgGray = styles.bgColor.bgBlackBright;
	styles.color.grey = styles.color.blackBright;
	styles.bgColor.bgGrey = styles.bgColor.bgBlackBright;

	for (const [groupName, group] of Object.entries(styles)) {
		for (const [styleName, style] of Object.entries(group)) {
			styles[styleName] = {
				open: `\u001B[${style[0]}m`,
				close: `\u001B[${style[1]}m`
			};

			group[styleName] = styles[styleName];

			codes.set(style[0], style[1]);
		}

		Object.defineProperty(styles, groupName, {
			value: group,
			enumerable: false
		});
	}

	Object.defineProperty(styles, 'codes', {
		value: codes,
		enumerable: false
	});

	styles.color.close = '\u001B[39m';
	styles.bgColor.close = '\u001B[49m';

	setLazyProperty(styles.color, 'ansi', () => makeDynamicStyles(wrapAnsi16, 'ansi16', ansi2ansi, false));
	setLazyProperty(styles.color, 'ansi256', () => makeDynamicStyles(wrapAnsi256, 'ansi256', ansi2ansi, false));
	setLazyProperty(styles.color, 'ansi16m', () => makeDynamicStyles(wrapAnsi16m, 'rgb', rgb2rgb, false));
	setLazyProperty(styles.bgColor, 'ansi', () => makeDynamicStyles(wrapAnsi16, 'ansi16', ansi2ansi, true));
	setLazyProperty(styles.bgColor, 'ansi256', () => makeDynamicStyles(wrapAnsi256, 'ansi256', ansi2ansi, true));
	setLazyProperty(styles.bgColor, 'ansi16m', () => makeDynamicStyles(wrapAnsi16m, 'rgb', rgb2rgb, true));

	return styles;
}

// Make the export immutable
Object.defineProperty(module, 'exports', {
	enumerable: true,
	get: assembleStyles
});
});

var browser = {
	stdout: false,
	stderr: false
};

const stringReplaceAll$1 = (string, substring, replacer) => {
	let index = string.indexOf(substring);
	if (index === -1) {
		return string;
	}

	const substringLength = substring.length;
	let endIndex = 0;
	let returnValue = '';
	do {
		returnValue += string.substr(endIndex, index - endIndex) + substring + replacer;
		endIndex = index + substringLength;
		index = string.indexOf(substring, endIndex);
	} while (index !== -1);

	returnValue += string.substr(endIndex);
	return returnValue;
};

const stringEncaseCRLFWithFirstIndex$1 = (string, prefix, postfix, index) => {
	let endIndex = 0;
	let returnValue = '';
	do {
		const gotCR = string[index - 1] === '\r';
		returnValue += string.substr(endIndex, (gotCR ? index - 1 : index) - endIndex) + prefix + (gotCR ? '\r\n' : '\n') + postfix;
		endIndex = index + 1;
		index = string.indexOf('\n', endIndex);
	} while (index !== -1);

	returnValue += string.substr(endIndex);
	return returnValue;
};

var util = {
	stringReplaceAll: stringReplaceAll$1,
	stringEncaseCRLFWithFirstIndex: stringEncaseCRLFWithFirstIndex$1
};

const TEMPLATE_REGEX = /(?:\\(u(?:[a-f\d]{4}|\{[a-f\d]{1,6}\})|x[a-f\d]{2}|.))|(?:\{(~)?(\w+(?:\([^)]*\))?(?:\.\w+(?:\([^)]*\))?)*)(?:[ \t]|(?=\r?\n)))|(\})|((?:.|[\r\n\f])+?)/gi;
const STYLE_REGEX = /(?:^|\.)(\w+)(?:\(([^)]*)\))?/g;
const STRING_REGEX = /^(['"])((?:\\.|(?!\1)[^\\])*)\1$/;
const ESCAPE_REGEX = /\\(u(?:[a-f\d]{4}|{[a-f\d]{1,6}})|x[a-f\d]{2}|.)|([^\\])/gi;

const ESCAPES = new Map([
	['n', '\n'],
	['r', '\r'],
	['t', '\t'],
	['b', '\b'],
	['f', '\f'],
	['v', '\v'],
	['0', '\0'],
	['\\', '\\'],
	['e', '\u001B'],
	['a', '\u0007']
]);

function unescape(c) {
	const u = c[0] === 'u';
	const bracket = c[1] === '{';

	if ((u && !bracket && c.length === 5) || (c[0] === 'x' && c.length === 3)) {
		return String.fromCharCode(parseInt(c.slice(1), 16));
	}

	if (u && bracket) {
		return String.fromCodePoint(parseInt(c.slice(2, -1), 16));
	}

	return ESCAPES.get(c) || c;
}

function parseArguments(name, arguments_) {
	const results = [];
	const chunks = arguments_.trim().split(/\s*,\s*/g);
	let matches;

	for (const chunk of chunks) {
		const number = Number(chunk);
		if (!Number.isNaN(number)) {
			results.push(number);
		} else if ((matches = chunk.match(STRING_REGEX))) {
			results.push(matches[2].replace(ESCAPE_REGEX, (m, escape, character) => escape ? unescape(escape) : character));
		} else {
			throw new Error(`Invalid Chalk template style argument: ${chunk} (in style '${name}')`);
		}
	}

	return results;
}

function parseStyle(style) {
	STYLE_REGEX.lastIndex = 0;

	const results = [];
	let matches;

	while ((matches = STYLE_REGEX.exec(style)) !== null) {
		const name = matches[1];

		if (matches[2]) {
			const args = parseArguments(name, matches[2]);
			results.push([name].concat(args));
		} else {
			results.push([name]);
		}
	}

	return results;
}

function buildStyle(chalk, styles) {
	const enabled = {};

	for (const layer of styles) {
		for (const style of layer.styles) {
			enabled[style[0]] = layer.inverse ? null : style.slice(1);
		}
	}

	let current = chalk;
	for (const [styleName, styles] of Object.entries(enabled)) {
		if (!Array.isArray(styles)) {
			continue;
		}

		if (!(styleName in current)) {
			throw new Error(`Unknown Chalk style: ${styleName}`);
		}

		current = styles.length > 0 ? current[styleName](...styles) : current[styleName];
	}

	return current;
}

var templates = (chalk, temporary) => {
	const styles = [];
	const chunks = [];
	let chunk = [];

	// eslint-disable-next-line max-params
	temporary.replace(TEMPLATE_REGEX, (m, escapeCharacter, inverse, style, close, character) => {
		if (escapeCharacter) {
			chunk.push(unescape(escapeCharacter));
		} else if (style) {
			const string = chunk.join('');
			chunk = [];
			chunks.push(styles.length === 0 ? string : buildStyle(chalk, styles)(string));
			styles.push({inverse, styles: parseStyle(style)});
		} else if (close) {
			if (styles.length === 0) {
				throw new Error('Found extraneous } in Chalk template literal');
			}

			chunks.push(buildStyle(chalk, styles)(chunk.join('')));
			chunk = [];
			styles.pop();
		} else {
			chunk.push(character);
		}
	});

	chunks.push(chunk.join(''));

	if (styles.length > 0) {
		const errMessage = `Chalk template literal is missing ${styles.length} closing bracket${styles.length === 1 ? '' : 's'} (\`}\`)`;
		throw new Error(errMessage);
	}

	return chunks.join('');
};

const {stdout: stdoutColor, stderr: stderrColor} = browser;
const {
	stringReplaceAll,
	stringEncaseCRLFWithFirstIndex
} = util;

const {isArray} = Array;

// `supportsColor.level` → `ansiStyles.color[name]` mapping
const levelMapping = [
	'ansi',
	'ansi',
	'ansi256',
	'ansi16m'
];

const styles = Object.create(null);

const applyOptions = (object, options = {}) => {
	if (options.level && !(Number.isInteger(options.level) && options.level >= 0 && options.level <= 3)) {
		throw new Error('The `level` option should be an integer from 0 to 3');
	}

	// Detect level if not set manually
	const colorLevel = stdoutColor ? stdoutColor.level : 0;
	object.level = options.level === undefined ? colorLevel : options.level;
};

class ChalkClass {
	constructor(options) {
		// eslint-disable-next-line no-constructor-return
		return chalkFactory(options);
	}
}

const chalkFactory = options => {
	const chalk = {};
	applyOptions(chalk, options);

	chalk.template = (...arguments_) => chalkTag(chalk.template, ...arguments_);

	Object.setPrototypeOf(chalk, Chalk.prototype);
	Object.setPrototypeOf(chalk.template, chalk);

	chalk.template.constructor = () => {
		throw new Error('`chalk.constructor()` is deprecated. Use `new chalk.Instance()` instead.');
	};

	chalk.template.Instance = ChalkClass;

	return chalk.template;
};

function Chalk(options) {
	return chalkFactory(options);
}

for (const [styleName, style] of Object.entries(ansiStyles)) {
	styles[styleName] = {
		get() {
			const builder = createBuilder(this, createStyler(style.open, style.close, this._styler), this._isEmpty);
			Object.defineProperty(this, styleName, {value: builder});
			return builder;
		}
	};
}

styles.visible = {
	get() {
		const builder = createBuilder(this, this._styler, true);
		Object.defineProperty(this, 'visible', {value: builder});
		return builder;
	}
};

const usedModels = ['rgb', 'hex', 'keyword', 'hsl', 'hsv', 'hwb', 'ansi', 'ansi256'];

for (const model of usedModels) {
	styles[model] = {
		get() {
			const {level} = this;
			return function (...arguments_) {
				const styler = createStyler(ansiStyles.color[levelMapping[level]][model](...arguments_), ansiStyles.color.close, this._styler);
				return createBuilder(this, styler, this._isEmpty);
			};
		}
	};
}

for (const model of usedModels) {
	const bgModel = 'bg' + model[0].toUpperCase() + model.slice(1);
	styles[bgModel] = {
		get() {
			const {level} = this;
			return function (...arguments_) {
				const styler = createStyler(ansiStyles.bgColor[levelMapping[level]][model](...arguments_), ansiStyles.bgColor.close, this._styler);
				return createBuilder(this, styler, this._isEmpty);
			};
		}
	};
}

const proto = Object.defineProperties(() => {}, {
	...styles,
	level: {
		enumerable: true,
		get() {
			return this._generator.level;
		},
		set(level) {
			this._generator.level = level;
		}
	}
});

const createStyler = (open, close, parent) => {
	let openAll;
	let closeAll;
	if (parent === undefined) {
		openAll = open;
		closeAll = close;
	} else {
		openAll = parent.openAll + open;
		closeAll = close + parent.closeAll;
	}

	return {
		open,
		close,
		openAll,
		closeAll,
		parent
	};
};

const createBuilder = (self, _styler, _isEmpty) => {
	const builder = (...arguments_) => {
		if (isArray(arguments_[0]) && isArray(arguments_[0].raw)) {
			// Called as a template literal, for example: chalk.red`2 + 3 = {bold ${2+3}}`
			return applyStyle(builder, chalkTag(builder, ...arguments_));
		}

		// Single argument is hot path, implicit coercion is faster than anything
		// eslint-disable-next-line no-implicit-coercion
		return applyStyle(builder, (arguments_.length === 1) ? ('' + arguments_[0]) : arguments_.join(' '));
	};

	// We alter the prototype because we must return a function, but there is
	// no way to create a function with a different prototype
	Object.setPrototypeOf(builder, proto);

	builder._generator = self;
	builder._styler = _styler;
	builder._isEmpty = _isEmpty;

	return builder;
};

const applyStyle = (self, string) => {
	if (self.level <= 0 || !string) {
		return self._isEmpty ? '' : string;
	}

	let styler = self._styler;

	if (styler === undefined) {
		return string;
	}

	const {openAll, closeAll} = styler;
	if (string.indexOf('\u001B') !== -1) {
		while (styler !== undefined) {
			// Replace any instances already present with a re-opening code
			// otherwise only the part of the string until said closing code
			// will be colored, and the rest will simply be 'plain'.
			string = stringReplaceAll(string, styler.close, styler.open);

			styler = styler.parent;
		}
	}

	// We can move both next actions out of loop, because remaining actions in loop won't have
	// any/visible effect on parts we add here. Close the styling before a linebreak and reopen
	// after next line to fix a bleed issue on macOS: https://github.com/chalk/chalk/pull/92
	const lfIndex = string.indexOf('\n');
	if (lfIndex !== -1) {
		string = stringEncaseCRLFWithFirstIndex(string, closeAll, openAll, lfIndex);
	}

	return openAll + string + closeAll;
};

let template;
const chalkTag = (chalk, ...strings) => {
	const [firstString] = strings;

	if (!isArray(firstString) || !isArray(firstString.raw)) {
		// If chalk() was called by itself or with a string,
		// return the string itself as a string.
		return strings.join(' ');
	}

	const arguments_ = strings.slice(1);
	const parts = [firstString.raw[0]];

	for (let i = 1; i < firstString.length; i++) {
		parts.push(
			String(arguments_[i - 1]).replace(/[{}\\]/g, '\\$&'),
			String(firstString.raw[i])
		);
	}

	if (template === undefined) {
		template = templates;
	}

	return template(chalk, parts.join(''));
};

Object.defineProperties(Chalk.prototype, styles);

const chalk = Chalk(); // eslint-disable-line new-cap
chalk.supportsColor = stdoutColor;
chalk.stderr = Chalk({level: stderrColor ? stderrColor.level : 0}); // eslint-disable-line new-cap
chalk.stderr.supportsColor = stderrColor;

var source = chalk;

var prettyHash = function prettyHash (buf) {
  if (Buffer.isBuffer(buf)) buf = buf.toString('hex');
  if (typeof buf === 'string' && buf.length > 8) {
    return buf.slice(0, 6) + '..' + buf.slice(-2)
  }
  return buf
};

const BYTE_UNITS = [
	'B',
	'kB',
	'MB',
	'GB',
	'TB',
	'PB',
	'EB',
	'ZB',
	'YB'
];

const BIBYTE_UNITS = [
	'B',
	'kiB',
	'MiB',
	'GiB',
	'TiB',
	'PiB',
	'EiB',
	'ZiB',
	'YiB'
];

const BIT_UNITS = [
	'b',
	'kbit',
	'Mbit',
	'Gbit',
	'Tbit',
	'Pbit',
	'Ebit',
	'Zbit',
	'Ybit'
];

const BIBIT_UNITS = [
	'b',
	'kibit',
	'Mibit',
	'Gibit',
	'Tibit',
	'Pibit',
	'Eibit',
	'Zibit',
	'Yibit'
];

/*
Formats the given number using `Number#toLocaleString`.
- If locale is a string, the value is expected to be a locale-key (for example: `de`).
- If locale is true, the system default locale is used for translation.
- If no value for locale is specified, the number is returned unmodified.
*/
const toLocaleString = (number, locale, options) => {
	let result = number;
	if (typeof locale === 'string' || Array.isArray(locale)) {
		result = number.toLocaleString(locale, options);
	} else if (locale === true || options !== undefined) {
		result = number.toLocaleString(undefined, options);
	}

	return result;
};

var prettyBytes = (number, options) => {
	if (!Number.isFinite(number)) {
		throw new TypeError(`Expected a finite number, got ${typeof number}: ${number}`);
	}

	options = Object.assign({bits: false, binary: false}, options);

	const UNITS = options.bits ?
		(options.binary ? BIBIT_UNITS : BIT_UNITS) :
		(options.binary ? BIBYTE_UNITS : BYTE_UNITS);

	if (options.signed && number === 0) {
		return ` 0 ${UNITS[0]}`;
	}

	const isNegative = number < 0;
	const prefix = isNegative ? '-' : (options.signed ? '+' : '');

	if (isNegative) {
		number = -number;
	}

	let localeOptions;

	if (options.minimumFractionDigits !== undefined) {
		localeOptions = {minimumFractionDigits: options.minimumFractionDigits};
	}

	if (options.maximumFractionDigits !== undefined) {
		localeOptions = Object.assign({maximumFractionDigits: options.maximumFractionDigits}, localeOptions);
	}

	if (number < 1) {
		const numberString = toLocaleString(number, options.locale, localeOptions);
		return prefix + numberString + ' ' + UNITS[0];
	}

	const exponent = Math.min(Math.floor(options.binary ? Math.log(number) / Math.log(1024) : Math.log10(number) / 3), UNITS.length - 1);
	// eslint-disable-next-line unicorn/prefer-exponentiation-operator
	number /= Math.pow(options.binary ? 1024 : 1000, exponent);

	if (!localeOptions) {
		number = number.toPrecision(3);
	}

	const numberString = toLocaleString(Number(number), options.locale, localeOptions);

	const unit = UNITS[exponent];

	return prefix + numberString + ' ' + unit;
};

function getLogLevel () {
  {
    return window.localStorage.getItem('LOG') || 'info'
  }
}

var log = function createLogger (opts) {
  const defaultOpts = {
    level: getLogLevel(),
    prettifier: createPrettifier,
    browser: {
      write: browserWrite
    }
  };

  defaultOpts.prettyPrint = {};

  const logger = browser$1({
    ...defaultOpts,
    ...opts
  });
  return logger
};

function browserWrite (obj) {
  let method = convertLogNumber(obj).name;
  if (method === 'fatal') method = 'error';
  if (method === 'trace') method = 'debug';
  if (!console[method]) method = 'log';
  const prettified = prettify(obj);
  console[method](prettified);
}

function createPrettifier () {
  return prettify
}

function prettify (obj) {
  if (typeof obj === 'string') {
    obj = { message: obj };
  }
  if (!obj.message && obj.msg) obj.message = obj.msg;
  if (obj.err) obj.error = obj.err;
  const sp = ' ';

  const date = formatDate();
  const level = formatLevel(obj);

  if (obj.error && typeof obj.error === 'object' && obj.error.message) {
    if (obj.message) obj.message += source.gray(' (' + obj.error.message + ')');
    else obj.message = obj.error.message;
  }

  let out = '[';
  if (date) out += date + sp;
  out += level;
  if (obj.name) out += sp + source.gray(obj.name);
  if (obj.namespace) out += sp + source.gray(obj.namespace);
  if (obj.collection) out += sp + source.blue(collectionLabel(obj.collection));
  out += '] ';
  if (obj.res && obj.req) out += formatHttp(obj) + sp;
  out += obj.message;
  if (obj.record) out += sp + formatRecord(obj.record);
  return out + '\n'
}

function formatDate () {
  const date = new Date();
  const z = date.getTimezoneOffset() * 60 * 1000;
  const dateLocal = new Date(date - z);
  return dateLocal.toISOString().substring(0, 19)
}

function formatRecord (record) {
  let out = '';
  out += source.blue.bold(record.id);
  out += ' ' + source.blue(record.type);
  const seq = record.seq || '?';
  out += ' ' + source.gray(`${prettyHash(record.key)}@${seq}`);
  return source.blue('Record(') + out + source.blue(')')
}

function collectionLabel (collection) {
  return collection._keyOrName || collection.name || (collection.key && prettyHash(collection.key)) || '<unknown>'
}

function formatLevel (obj) {
  const { name, color } = convertLogNumber(obj);
  let format = source[color];
  if (obj.level >= 40) format = source[color].bold;
  return format(name.toUpperCase())
}

function formatHttp (obj) {
  if (!obj.res) return
  const { res, req } = obj;
  let color;
  if (res.statusCode > 200) color = 'red';
  else color = 'green';
  let out = '';
  out += source.bold(req.method);
  out += ' ' + req.url;
  out += ' ' + source.gray(obj.responseTime + 'ms');
  out += ' ' + source[color].bold(res.statusCode);
  const size = res.headers['content-length'];
  if (size) out += ' ' + source.gray(prettyBytes(Number(size)).replace(' ', ''));
  return out
}

function convertLogNumber (obj) {
  if (obj.level === 10) return { name: 'trace', color: 'gray' }
  if (obj.level === 20) return { name: 'debug', color: 'gray' }
  if (obj.level === 30) return { name: 'info', color: 'green' }
  if (obj.level === 40) return { name: 'warn', color: 'yellow' }
  if (obj.level === 50) return { name: 'error', color: 'red' }
  if (obj.level === 60) return { name: 'fatal', color: 'red' }
}

const {
  DEFAULT_ENDPOINT,
  DEFAULT_WORKSPACE
} = constants;

function defaultWorkspace () {
  return (process && {} && {}.WORKSPACE) || DEFAULT_WORKSPACE
}

class Workspace {
  /**
   * The manager for a remote connection to a Sonar workspace.
   *
   * @constructor
   * @param {object} [opts] - Optional options.
   * @param {string} [opts.endpoint=http://localhost:9191/api] - The API endpoint to talk to.
   * @param {string} [opts.accessCode] - An access code to login at the endpoint.
   * @param {string} [opts.token] - A JSON web token to authorize to the endpoint.
   * @param {string} [opts.name] - The name of this client.
   */
  constructor (opts = {}) {
    this._workspace = opts.workspace || defaultWorkspace();
    this.endpoint = opts.endpoint || DEFAULT_ENDPOINT;
    if (this.endpoint.endsWith('/')) {
      this.endpoint = this.endpoint.substring(0, this.endpoint.length - 1);
    }
    this.endpoint = this.endpoint + '/workspace/' + this._workspace;
    this._collections = new Map();
    this._id = opts.id || randomBytes(16).toString('hex');
    this._token = opts.token;
    this._accessCode = opts.accessCode;
    this._eventSources = [];

    this.log = opts.log || new log();
    this.bots = new bots(this);
  }

  /**
   * Closes the client.
   *
   * @async
   * @return {Promise<void>}
   */
  async close () {
    for (const eventSource of this._eventSources) {
      eventSource.close();
    }
    for (const collection of this._collections.values()) {
      collection.close();
    }
  }

  async open () {
    if (this.opened) return
    if (!this._openPromise) this._openPromise = this._open();
    await this._openPromise;
  }

  async _open () {
    await this._login();
    this.opened = true;
  }

  // TODO: Support re-logins
  async _login () {
    if (this._accessCode && !this._token) {
      const res = await this.fetch('/login', { params: { code: this._accessCode }, method: 'POST', opening: true });
      const token = res.token;
      this._token = token;
    }
  }

  /**
   * Get a list of all collections available on this endpoint.
   *
   * @async
   * @return {Promise.<object[]>} Promise that resolves to an array of collection info objects.
   */
  async listCollections () {
    const info = await this.fetch('/info');
    return info.collections
  }

  /**
   * Creates a collection with name name on the Sonar server. The name may not contain whitespaces. opts is an optional object with:
   *
   * @async
   * @param {string} name - Name of the new collection, may not contain whitespaces.
   * @param {object} [opts] - Optional options object.
   * @param {string} [opts.key] - Hex string of an existing collection. Will then sync this collection instead of creating a new, empty collection.
   * @param {string} [opts.alias] - When setting key, alias is required and is your nick name within this collection.
   * @return {Promise<Collection>} The created collection.
   */
  async createCollection (name, opts = {}) {
    opts.name = name;
    await this.fetch('/collection', {
      method: 'POST',
      body: opts
    });
    return this.openCollection(name)
  }

  // TODO: Move to Collection.update()?
  // TODO: info == config?
  /**
   * Updates the config of a collection.
   *
   * @async
   * @param {string} name - Name of the collection.
   * @param {object} info - [TODO:description]
   * @param {boolean} info.share - Controls whether a collection is shared via p2p.
   * @return {Promise<void>}
   */
  async updateCollection (name, info) {
    return this.fetch('/collection/' + name, {
      method: 'PATCH',
      body: info
    })
  }

  /**
   * Returns a Collection object for a given key or name of a collection.
   *
   * @async
   * @param {string} keyOrName - Key or name of the collection to open/return.
   * @return {Promise<Collection>}
   */
  async openCollection (keyOrName) {
    if (this._collections.has(keyOrName)) {
      const collection = this._collections.get(keyOrName);
      if (!collection.opened) await collection.open();
      return collection
    }

    const collection$1 = new collection(this, keyOrName);
    this._collections.set(keyOrName, collection$1);
    // This will throw if the collection does not exist.
    await collection$1.open();
    this._collections.set(collection$1.name, collection$1);
    this._collections.set(collection$1.key, collection$1);
    return collection$1
  }

  getAuthHeaders (opts = {}) {
    const headers = {};
    const token = this._token || opts.token;
    if (token) {
      headers.authorization = 'Bearer ' + token;
    }
    return headers
  }

  getHeaders (opts = {}) {
    return this.getAuthHeaders()
  }

  /**
   * Fetch a resource.
   *
   * This is a wrapper around the fetch web API. It should be API compatible to fetch,
   * with the following changes:
   *
   * @async
   * @param {string} [opts.requestType='json'] Request encoding and content type.
   *   Supported values are 'json' and 'binary'
   * @param {string} [opts.responseType='text'] Response encoding. If the response
   *    has a JSON content type, will always be set to 'json'.
   *    Supported values are 'text', 'binary' and 'stream'.
   * @param {object} [opts.params] Query string parameters (will be encoded correctly).
   *
   * @return {Promise<object>} If the response has a JSON content type header, the
   *    decoded JSON will be returned. if opts.responseType is 'binary' or 'text',
   *    the response will be returned as a buffer or text.
   *
   * TODO: Rethink the default responseType cascade.
   */
  async fetch (url, opts = {}) {
    if (!this.opened && !opts.opening) {
      await this.open();
    }
    if (!opts.endpoint && this.endpoint) {
      opts.endpoint = this.endpoint;
    }
    if (!opts.headers) opts.headers = {};
    opts.headers = {
      ...opts.headers,
      ...this.getAuthHeaders(opts)
    };
    opts.log = message => this.log.debug({ message, name: 'fetch' });
    return fetch_1(url, opts)
  }

  createEventSource (path, opts = {}) {
    if (!opts.endpoint && this.endpoint) opts.endpoint = this.endpoint;
    opts.headers = Object.assign(opts.headers || {}, this.getHeaders(opts));
    const url = (opts.endpoint || '') + path;
    const eventSource = new EventSource(url, opts);
    eventSource.addEventListener('message', message => {
      try {
        const event = JSON.parse(message.data);
        if (opts.onmessage) opts.onmessage(event);
      } catch (e) {}
    });
    eventSource.addEventListener('error', err => {
      // TODO: Where do these errors go?
      // TODO: After a couple of fails die.
      if (opts.onerror) opts.onerror(err);
      else console.error('Event source error', err);
    });
    this._eventSources.push(eventSource);
    return eventSource
  }
}

var workspace = Workspace;

var searchquerybuilder = class SearchQueryBuilder {
  constructor (schema) {
    this.schema = schema;
    this.query = {};
  }

  bool (boolType, queries) {
    if (!this.query.hasOwnProperty('bool')) {
      this.query.bool = {};
    }
    this.query.bool[boolType] = queries;
    return this
  }

  phrase (field, terms) {
    if (!this.query.hasOwnProperty('phrase')) {
      this.query.phrase = {};
    }
    this.query.phrase[field] = { terms };
    return this
  }

  limit (limit) {
    this.limit = limit;
    return this
  }

  term (field, value) {
    return { term: { [field]: value } }
  }

  getQuery () {
    return { query: this.query, limit: this.limit }
  }
};

const { DEFAULT_COLLECTION } = constants;


var legacyClient = class LegacyClient extends workspace {
  constructor (opts) {
    super(opts);
    this._focus = null;
    this._defaultCollection = opts.collection || DEFAULT_COLLECTION;
  }

  focusCollection (name) {
    this._focus = this.openCollection(name);
  }

  async focusedCollection () {
    if (!this._focus && this._defaultCollection) this.focusCollection(this._defaultCollection);
    this._focusedCollection = await this._focus;
    return this._focus
  }

  // TODO: Remove.
  async getCurrentCollection () {
    return this.focusedCollection()
  }

  async info () {
    return this.fetch('/info')
  }

  async getTypes () {
    const collection = await this.focusedCollection();
    return collection.schema.getTypes()
  }

  async getType (name) {
    const collection = await this.focusedCollection();
    return collection.schema.getType(name)
  }

  async putType (name, schema) {
    schema.name = name;
    const collection = await this.focusedCollection();
    return collection.putType(schema)
  }

  // TODO: Remove.
  getSchemas () {
    return this.getTypes()
  }

  getSchema (name) {
    return this.getType(name)
  }

  async putFeed (key, opts) {
    const collection = await this.focusedCollection();
    return collection.putFeed(key, opts)
  }

  async get (req, opts) {
    const collection = await this.focusedCollection();
    return collection.get(req, opts)
  }

  async put (record) {
    const collection = await this.focusedCollection();
    return collection.put(record)
  }

  async del (record) {
    const collection = await this.focusedCollection();
    return collection.del(record)
  }

  async sync (view) {
    const collection = await this.focusedCollection();
    return collection.sync(view)
  }

  async query (name, args, opts) {
    const collection = await this.focusedCollection();
    return collection.query(name, args, opts)
  }

  async search (args, opts) {
    if (typeof args === 'string') {
      args = JSON.stringify(args);
    } else if (args instanceof searchquerybuilder) {
      args = args.getQuery();
    }
    const collection = await this.focusedCollection();
    return collection.query('search', args, opts)
  }

  // FS methods
  async getDrives () {
    const collection = await this.focusedCollection();
    return collection.fs.listDrives()
  }

  async readdir (path) {
    const collection = await this.focusedCollection();
    return collection.fs.readdir(path)
  }

  async writeFile (path, file, opts) {
    const collection = await this.focusedCollection();
    return collection.fs.writeFile(path, file, opts)
  }

  async readFile (path, opts) {
    const collection = await this.focusedCollection();
    return collection.fs.readFile(path, opts)
  }

  async statFile (path) {
    const collection = await this.focusedCollection();
    return collection.fs.stat(path)
  }

  // Resources
  // TODO: Rethink the resource API and model
  async createResource (value, opts) {
    const collection = await this.focusedCollection();
    return collection.resources.create(value, opts)
  }

  async readResourceFile (record, opts) {
    const collection = await this.focusedCollection();
    return collection.resources.readFile(record, opts)
  }

  async writeResourceFile (record, file, opts) {
    const collection = await this.focusedCollection();
    return collection.resources.writeFile(record, file, opts)
  }

  async resolveResourceURL (record) {
    const collection = await this.focusedCollection();
    return collection.resources.resolveFileURL(record)
  }

  // Subscriptions
  async pullSubscription (name, opts) {
    const collection = await this.focusedCollection();
    return collection.pullSubscription(name, opts)
  }

  async ackSubscription (name, cursor) {
    const collection = await this.focusedCollection();
    return collection.ackSubscription(name, cursor)
  }
};

var client = legacyClient;
var Workspace_1 = workspace;
var SearchQueryBuilder_1 = searchquerybuilder;
client.Workspace = Workspace_1;
client.SearchQueryBuilder = SearchQueryBuilder_1;

export default client;
export { SearchQueryBuilder_1 as SearchQueryBuilder, Workspace_1 as Workspace };
//# sourceMappingURL=index.js.map
